#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""The ds2 module contains classes representing DS2 code constructs."""

import re
import uuid
from collections import namedtuple, OrderedDict


from ..decorators import deprecated, versionadded


@deprecated('Use DS2PyMASPackage instead.', version='1.5', removed_in='1.6')
class DS2Package(object):  # skipcq PYL-R0205
    def __init__(
        self, variables, code=None, return_code=True, return_message=True, target=None
    ):
        self._id = uuid.uuid4().hex.upper()
        self._python_code = code or []
        code = code or []

        self.methods = [
            DS2PyMASMethod(
                self._id, variables, code, return_code, return_message, target
            )
        ]

        self._body = (
            "dcl package pymas py;",
            "dcl package logger logr('App.tk.MAS');",
            "dcl varchar(67108864) character set utf8 pycode;",
            "dcl int revision;",
        )

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        # Max length in some SAS products in 32 characters
        return ('_' + str(self.id))[:32]

    def code(self):
        code = (
            ("package %s / overwrite=yes;" % self.name,)
            + tuple('    ' + line for line in self._body)
            + ('',)
        )

        for method in self.methods:
            code += tuple('    ' + line for line in method.code().split('\n'))
        code += ("endpackage;", '')

        return '\n'.join(code)


@versionadded(version='1.5')
class DS2BasePackage(object):  # skipcq PYL-R0205
    """Defines a DS2 package.

    Parameters
    ----------
    code : str
        Any code to include in the body of the package.

    Attributes
    ----------
    methods : list
        A collection of :class:`DS2BaseMethod` instances that will be included
        in the package definition.

    """

    def __init__(self, code=None):
        self._id = uuid.uuid4().hex.upper()
        self._body = code or ()
        self.methods = []

    @property
    def id(self):
        """Unique identifier generated for the package."""
        return self._id

    @property
    def name(self):
        """Unique name generated for the package."""
        # Max length in some SAS products in 32 characters
        return ('_' + str(self.id))[:32]

    def code(self):
        """Get the DS2 code for the package.

        Returns
        -------
        str

        """
        code = (
            ("package %s / overwrite=yes;" % self.name,)
            + tuple('    ' + line for line in self._body)
            + ('',)
        )

        for method in self.methods:
            code += tuple('    ' + line for line in method.code().split('\n'))
        code += ("endpackage;", '')

        return '\n'.join(code)


@versionadded(version='1.5')
class DS2PyMASPackage(DS2BasePackage):
    """A DS2 package that uses PyMAS to invoke Python code.

    Parameters
    ----------
    code

    """

    def __init__(self, code=None):
        body = (
            "dcl package pymas py;",
            "dcl package logger logr('App.tk.MAS');",
            "dcl varchar(67108864) character set utf8 pycode;",
            "dcl int revision;",
        )

        super(DS2PyMASPackage, self).__init__(body)

        self._python_code = code or []
        code = code or []

        self.methods.append(
            # Package init() method
            DS2PyMASMethod(
                self._id,
                [],
                code,
                return_code=False,
                return_message=False,
                target=None,
                method_name='init',
            )
        )

    def add_method(self, name, target, variables):
        """Add a DS2 method that calls a Python function defined by the package.

        Parameters
        ----------
        name : str
            Name of the DS2 method to create.
        target : str
            Name of the Python method to call
        variables : list of :class:`DS2Variable`
            List of input and output variables for the method.

        Returns
        -------
        None

        """
        public_variables = list(variables)
        private_variables = []

        # Add a return code if not already present
        if not any(v for v in public_variables if v.name.lower() == 'rc'):
            private_variables.append(DS2Variable('rc', 'int', True))

        body = [v.as_declaration() for v in private_variables]

        body += [
            "dcl varchar(4068) msg;",
            "rc = py.useMethod('%s');" % target,
            "if rc then return;",
        ]

        # Set Python input variables
        body += [
            '%s    if rc then return;' % v.pymas_statement()
            for v in public_variables
            if not v.out
        ]

        # Execute Python method
        body += ['rc = py.execute();    if rc then return;']

        # Get Python output variables
        body += [
            v.pymas_statement() for v in public_variables if v.out and v.name != 'rc'
        ]

        # Log any error messages returned
        body += [
            "msg = py.getString('msg');",
            'if not null(msg) then logr.log(\'e\', \'Error executing '
            'Python function "%s": $s\', msg);' % name,
        ]

        self.methods.append(DS2BaseMethod(name, variables, body))


class DS2BaseMethod(object):  # skipcq PYL-R0205
    def __init__(self, name, variables, body=None):
        self._name = name
        self.variables = variables

        if body is None:
            self._body = []
        elif isinstance(body, str):
            self._body = body.split('\n')
        else:
            self._body = list(body)

    @property
    def name(self):
        return self._name

    def code(self):
        vars = ',\n'.join('    %s' % v.as_parameter() for v in self.variables)

        # Don't spread signature over multiple lines if there are no variables
        if vars:
            func = ('method %s(' % self.name, vars, '    );', '')
        else:
            func = ('method %s();' % self.name, '')

        if self._body:
            func += tuple('    ' + line for line in self._body)

        func += ('end;', '')

        return '\n'.join(func)


class DS2PyMASMethod(DS2BaseMethod):
    """A DS2 method that builds a PyMAS object.

    Parameters
    ----------
    name : str
        Name of Python model to define
    variables
    python_code
    return_code
    return_message
    target
    method_name
    """

    def __init__(
        self,
        name,
        variables,
        python_code,
        return_code=None,
        return_message=None,
        target='wrapper',
        method_name='score',
    ):

        # target = target or 'wrapper'

        if isinstance(python_code, str):
            python_code = python_code.split('\n')

        self.public_variables = variables
        self.private_variables = []

        if return_code:
            self.public_variables.append(DS2Variable('rc', 'int', True))
        else:
            self.private_variables.append(DS2Variable('rc', 'int', True))

        if return_message:
            self.public_variables.append(DS2Variable('msg', 'char', True))

        body = [v.as_declaration() for v in self.private_variables]

        body += [
            "if null(py) then do;",
            "    py = _new_ pymas();",
            "    rc = py.useModule('%s', 1);" % name,
            "    if rc then do;",
        ]

        body += ["        rc = py.appendSrcLine('%s');" % l for l in python_code]
        body += [
            "        pycode = py.getSource();",
            "        revision = py.publish(pycode, '%s');" % name,
            "        if revision lt 1 then do;",
            "            logr.log('e', 'py.publish() failed.');",
            "            rc = -1;",
            "        end;",
            "    end;",
        ]
        if target is not None:
            body += ["    rc = py.useMethod('%s');" % target, "    if rc then return;"]
        body += ["end;"]

        if target is not None:
            # Set Python input variables
            body += [
                '%s    if rc then return;' % v.pymas_statement()
                for v in self.public_variables
                if not v.out
            ]

            # Execute Python method
            body += ['rc = py.execute();    if rc then return;']

            # Get Python output variables
            body += [
                v.pymas_statement()
                for v in self.public_variables
                if v.out and v.name != 'rc'
            ]

        super(DS2PyMASMethod, self).__init__(method_name, variables, body=body)


@deprecated(version='1.5', removed_in='1.6')
class DS2ScoreMethod(DS2BaseMethod):
    def __init__(
        self,
        variables,
        return_code=True,
        return_message=True,
        target='wrapper',
    ):
        self._target = target

        self.public_variables = variables
        self.private_variables = []

        if return_code:
            self.public_variables.append(DS2Variable('rc', 'int', True))
        else:
            self.private_variables.append(DS2Variable('rc', 'int', True))

        if return_message:
            self.public_variables.append(DS2Variable('msg', 'char', True))
        else:
            self.private_variables.append(DS2Variable('msg', 'char', True))

        body_statements = (
            [v.as_declaration() for v in self.private_variables]
            + [
                v.pymas_statement() + '    if rc then return;'
                for v in self.public_variables
                if not v.out
            ]
            + ['rc = py.execute();    if rc then return;']
            + [
                v.pymas_statement()
                for v in self.public_variables
                if v.out and v.name != 'rc'
            ]
        )

        super(DS2ScoreMethod, self).__init__('score', variables, body=body_statements)


# @versionadded(version='1.5')
# class DS2PredictProbaMethod(DS2BaseMethod):
#     def __init__(self, variables):
#
#         self.public_variables = variables
#         self.private_variables = []
#
#         body_statements = []
#
#         super(DS2PredictProbaMethod, self).__init__('predict_proba',
#         variables,
#                                                     body=body_statements)


@deprecated(version='1.5', removed_in='1.6')
class DS2Method(object):  # skipcq PYL-R0205
    def __init__(self, variables, code, target='wrapper'):
        self.variables = variables
        self._code = code
        self.target = target

    @property
    def name(self):
        return 'score'

    def code(self, return_code=True, return_message=True):
        signature = [v.as_parameter() for v in self.variables]

        if return_code:
            signature += ['in_out double rc']
        if return_message:
            signature += ['in_out char msg']

        signature = ', '.join(signature)

        # Python code to be embedded passed to PyMAS
        code = [self._code] if isinstance(self._code, str) else self._code
        code = ["\t\t\t\t\terr = py.appendSrcLine('{}');".format(line) for line in code]
        code = '\n'.join(code)

        set_statements = [v.pymas_statement() for v in self.variables if not v.out]
        get_statements = [v.pymas_statement() for v in self.variables if v.out]

        if return_code:
            get_statements += ['rc = err;']
        if return_message:
            get_statements += ["msg = py.getString('msg');"]

        ds2_statements = set_statements + ['err = py.execute();'] + get_statements
        ds2_statements = '\n'.join('\t\t\t\t' + s for s in ds2_statements)

        func = (
            '            method {}({});'.format(self.name, signature),
            '                dcl double err;',
            '                if null(py) then do;',
            '                    py = _new_ pymas();',
            code,
            '                    pycode = py.getSource();',
            "                    revision = py.publish( pycode, 'mypymodule' );",
            "                    err = py.useMethod('{}');".format(self.target),
            "                    if err then return;",
            "                end;",
            ds2_statements,
            "            end;",
        )
        func = '\n'.join(func)

        return func


class DS2Thread(object):  # skipcq PYL-R0205
    def __init__(
        self,
        variables,
        table,
        column_names=None,
        return_code=True,
        return_message=True,
        package=None,
        method=None,
    ):

        self._id = uuid.uuid4().hex.upper()
        self.table = table
        self.variables = variables
        self.return_code = return_code
        self.return_message = return_message
        self.column_names = column_names
        self.package = package

        # Default to predict() method if present
        if method is None and any(m.name == 'predict' for m in package.methods):
            method = next(x for x in package.methods if x.name == 'predict')

        # Fall back to score() method if present
        if method is None and any(m.name == 'score' for m in package.methods):
            method = next(x for x in package.methods if x.name == 'score')

        # Assume first method is init(), so fall back to next method
        if method is None and len(package.methods) > 1:
            method = package.methods[1]

        self.method = method

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return 'pyMasThread'

    def __str__(self):
        array_input = any(v.is_array for v in self.variables if not v.out)

        # If passing column data into Python as an array, need extra assignment statements to set values
        if array_input and self.column_names is not None:
            array = next(filter(lambda v: v.is_array, self.variables))
            var_assignments = [
                '{}[{}] = {};'.format(array.name, i + 1, self.column_names[i])
                for i in range(min(array.size, len(self.column_names)))
            ]
            var_assignments = '\n'.join(var_assignments)
        else:
            var_assignments = ''

        # Declare output variables.  Input variables are assumed to be columns in the input table.
        declarations = '\n'.join(
            v.as_declaration() for v in self.variables if v.out or v.is_array
        )

        keep_vars = [v.name for v in self.variables]

        code = (
            "thread {} / inline;".format(self.name),
            "  dcl package {} pythonPackage();".format(self.package.name),
            declarations,
            "  method run();",
            "    set SASEP.in;",
            var_assignments,
            "    pythonPackage.init();",
            "    pythonPackage.{}({});".format(self.method.name, ','.join(keep_vars)),
            "    output;",
            "  end;",
            "endthread;",
        )

        code = '\n'.join(code)

        return code


class DS2Variable(namedtuple('Ds2Variable', ['name', 'type', 'out'])):
    PY_TYPE_TO_DS2 = OrderedDict(
        [
            ('double64', 'double'),
            ('double32', 'double'),
            ('double', 'double'),
            # Terminates search if full string matches
            ('float64', 'double'),
            ('float32', 'double'),
            ('float', 'double'),
            ('string', 'char'),
            ('str', 'char'),
            ('varchar', 'char'),
            ('integer64', 'integer'),
            ('integer32', 'integer'),
            ('integer', 'integer'),
            # Terminates search if full string matches
            ('int64', 'integer'),
            ('int32', 'integer'),
            ('int', 'integer'),
            ('uint8', 'integer'),
            ('uint16', 'integer'),
            ('uint32', 'integer'),
            ('uint64', 'integer'),
        ]
    )

    DS2_TYPE_TO_VIYA = OrderedDict(
        [('double', 'decimal'), ('varchar', 'string'), ('char', 'string')]
    )

    def __new__(cls, *args, **kwargs):

        # Convert Python types to DS2 types if necessary
        if 'type' in kwargs:
            kwargs['type'] = DS2Variable._map_type(cls.PY_TYPE_TO_DS2, kwargs['type'])
        elif len(args) > 1:
            args = list(args)
            args[1] = DS2Variable._map_type(cls.PY_TYPE_TO_DS2, args[1])

        return super(DS2Variable, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def _map_type(cls, mapping, t):
        # Convert Python type names to DS2 type names
        t = str(t).lower().strip()

        # Using replace since type could be an array: float[10]
        for k, v in mapping.items():
            if t.startswith(k):
                t = t.replace(k, v)
                break

        return t

    def as_model_metadata(self):
        viya_type = self._map_type(self.DS2_TYPE_TO_VIYA, self.type)
        role = 'Output' if self.out else 'input'

        return OrderedDict([('name', self.name), ('role', role), ('type', viya_type)])

    def as_declaration(self):
        """DS2 variable declaration statement."""
        match = re.search(r'\[\d+\]$', self.type)
        if match is None:
            return 'dcl {} {};'.format(self.type, self.name)

        # Type is an array
        return 'dcl {} {}{};'.format(
            self.type[: match.start()], self.name, self.type[match.start() :]
        )

    def as_parameter(self):
        """DS2 parameter syntax for method signatures."""
        match = re.search(r'\[\d+\]$', self.type)
        param = self.name

        param = (
            self.type + ' ' + param
            if match is None
            else self.type[: match.start() :] + ' ' + param + '[*]'
        )
        param = 'in_out ' + param if self.out else param

        return param

    @property
    def size(self):
        match = re.search(r'\[\d+\]$', self.type)
        if match:
            return int(self.type[match.start() + 1 : match.end() - 1])
        return 0

    def pymas_statement(self, python_var_name=None):
        """Returns appropriate PyMAS get/set statements to move values between
        a Python variable and the DS2 variable.

        Parameters
        ----------
        python_var_name : str
            Python variable name.

        Returns
        -------

        """

        # Default Python variable name to DS2 variable name
        python_var_name = python_var_name or self.name

        if self.out:
            if self.type.startswith('double'):
                if self.is_array:
                    return "py.getDoubleArray('{}', {}, ret);".format(
                        python_var_name, self.name
                    )
                return "{} = py.getDouble('{}');".format(self.name, python_var_name)
            if self.type.startswith('char'):
                if self.is_array:
                    return "py.getStringArray('{}', {}, ret);".format(
                        python_var_name, self.name
                    )
                return "{} = py.getString('{}');".format(self.name, python_var_name)
            if self.type.startswith('integer'):
                if self.is_array:
                    return "py.getIntArray('{}', {}, ret);".format(
                        python_var_name, self.name
                    )
                return "{} = py.getInt('{}');".format(self.name, python_var_name)

            raise ValueError("Can't generate a DS2 statement for type `%s`" % self.type)

        # If we got this far, it's an input variable
        if self.type.startswith('double'):
            if self.is_array:
                return "rc = py.setDoubleArray('{}', {});".format(
                    python_var_name, self.name
                )
            return "rc = py.setDouble('{}', {});".format(python_var_name, self.name)

        if self.type.startswith('char'):
            if self.is_array:
                return "rc = py.setStringArray('{}', {});".format(
                    python_var_name, self.name
                )
            return "rc = py.setString('{}', {});".format(python_var_name, self.name)

        if self.type.startswith('integer'):
            if self.is_array:
                return "rc = py.setIntArray('{}', {});".format(
                    python_var_name, self.name
                )
            return "rc = py.setInt('{}', {});".format(python_var_name, self.name)
        raise ValueError("Can't generate a DS2 statement for type `%s`" % self.type)

    @property
    def is_array(self):
        # If type contains [n] then it's an array
        return re.search(r'\[\d+\]$', self.type) is not None



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import concurrent.futures
import copy
import logging
import json
import netrc
import os
import re
import ssl
import sys
import warnings
from datetime import datetime, timedelta
from uuid import UUID, uuid4

import requests
import requests.exceptions
import yaml
from packaging import version
from requests.adapters import HTTPAdapter
from urllib.parse import urlsplit, urlunsplit
from urllib.error import HTTPError

try:
    import swat
except ImportError:
    swat = None

try:
    import kerberos
except ImportError:
    try:
        import winkerberos as kerberos
    except ImportError:
        kerberos = None

from .utils.cli import sasctl_command
from .utils.misc import versionadded
from . import exceptions

logger = logging.getLogger(__name__)

_session = None


def _pformat(text):
    from pprint import pformat

    try:
        return pformat(json.loads(text))
    except (TypeError, UnicodeDecodeError, ValueError):
        try:
            return pformat(text)
        except UnicodeDecodeError:
            return text


def _redact(pattern, repl, string):
    is_bytes = isinstance(string, bytes)

    try:
        string = string.decode('utf-8') if is_bytes else string
        string = re.sub(pattern, repl, string)
        string = string.encode('utf-8') if is_bytes else string
    except UnicodeDecodeError:
        pass
    return string


def _filter_password(r):
    if hasattr(r, 'body') and r.body is not None:
        # Filter password from 'grant_type=password&username=<user>&password=<password>' during Post to Logon service.
        r.body = _redact(r"(?<=&password=)([^&]*)\b", '*****', r.body)

        # Filter client secret {"client_secret": "<password>"}
        r.body = _redact('(?<=client_secret": ")[^"]*', '*****', r.body)
    return r


def _filter_token(r):
    # Redact "Bearer <token>" in Authorization headers
    if hasattr(r, 'headers') and 'Authorization' in r.headers:
        r.headers['Authorization'] = _redact(
            r'(?<=Bearer ).*', '[redacted]', r.headers['Authorization']
        )

    # Redact "Basic <base64 encoded pw> in Authorization headers.  This covers client ids & client secrets.
    if hasattr(r, 'headers') and 'Authorization' in r.headers:
        r.headers['Authorization'] = _redact(
            r'(?<=Basic ).*', '[redacted]', r.headers['Authorization']
        )

    # Redact Consul token from headers.  Should only appear when registering a new client
    if hasattr(r, 'headers') and 'X-Consul-Token' in r.headers:
        r.headers['X-Consul-Token'] = '[redacted]'

    # Redact "access_token":"<token>" in response from Logon service
    if hasattr(r, '_content'):
        r._content = _redact('(?<=access_token":")[^"]*', '[redacted]', r._content)

    return r


DEFAULT_FILTERS = [_filter_password, _filter_token]


def current_session(*args, **kwargs):
    """Gets and sets the current session.

    If call with no arguments, the current session instance is returned, or
    None if no session has been created yet.  If called with an existing session
    instance, that session will be set as the default.  Otherwise, a new
    `Session` is instantiated using the provided arguments and set as the
    default.

    Parameters
    ----------
    args : any
    kwargs : any

    Returns
    -------
    Session

    Examples
    --------

    Get the current session

    >>> current_session()
    <sasctl.core.Session object at 0x1393fc550>

    Clear the current session

    >>> current_session(None)


    Make a new session current

    >>> current_session('example.com', 'knight', 'Ni!')
    <sasctl.core.Session object at 0x15a9df491>

    """
    global _session  # skipcq PYL-W0603

    # Explicitly set or clear the current session
    if len(args) == 1 and (isinstance(args[0], Session) or args[0] is None):
        _session = args[0]
    # Create a new session
    elif args:
        _session = Session(*args, **kwargs)

    return _session


class OAuth2Token(requests.auth.AuthBase):
    def __init__(
        self,
        access_token,
        refresh_token=None,
        expiration=None,
        expires_in=None,
        **kwargs
    ):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expiration = expiration

        if expires_in is not None:
            self.expiration = datetime.now() + timedelta(seconds=expires_in)

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer ' + self.access_token
        return r

    @property
    def is_expired(self):
        if self.expiration is None:
            return False

        return self.expiration < datetime.now()


class RestObj(dict):
    def __getattr__(self, item):
        # Only called when __getattribute__ failed to find the attribute
        # Return the item from underlying dictionary if possible.
        if item in self:
            result = self[item]

            if isinstance(result, dict):
                return RestObj(result)

            return result

        raise AttributeError(
            "'%s' object has no attribute '%s'" % (self.__class__.__name__, item)
        )

    def __repr__(self):
        headers = getattr(self, '_headers', {})

        return "%s(headers=%r, data=%s)" % (
            self.__class__,
            headers,
            super(RestObj, self).__repr__(),
        )

    def __str__(self):
        if 'name' in self:
            return str(self['name'])
        if 'id' in self:
            return str(self['id'])
        return repr(self)


class SSLContextAdapter(HTTPAdapter):
    """HTTPAdapter that uses the default SSL context on the machine."""

    def __init__(self, *args, **kwargs):
        self.assert_hostname = kwargs.pop('assert_hostname', True)
        requests.adapters.HTTPAdapter.__init__(self, *args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        context = ssl.create_default_context()
        context.check_hostname = self.assert_hostname
        kwargs['ssl_context'] = context
        kwargs['assert_hostname'] = self.assert_hostname
        return super(SSLContextAdapter, self).init_poolmanager(*args, **kwargs)


class Session(requests.Session):
    """Establish a connection to a SAS Viya server.

    Parameters
    ----------
    hostname : str or swat.CAS
        Name of the server to connect to or an established swat.CAS session.
    username : str, optional
        Username for authentication.  Not required if `host` is a CAS
        connection, if Kerberos is used, or if `token` is provided.  If using Kerberos and an explicit
        username is desired, maybe be a string in 'user@REALM' format.
    password : str, optional
        Password for authentication.  Not required when `host` is a CAS
        connection, `authinfo` is provided, `token` is provided, or Kerberos is used.
    authinfo : str, optional
        Path to a .authinfo or .netrc file from which credentials should be
        pulled.
    protocol : {'http', 'https'}
        Whether to use HTTP or HTTPS connections.  Defaults to `https`.
    port : int, optional
        Port number for the connection if a non-standard port is used.
        Defaults to 80 or 443 depending on `protocol`.
    verify_ssl : bool, optional
        Whether server-side SSL certificates should be verified.  Defaults
        to true.  Ignored for HTTP connections.
    token : str, optional
        OAuth token to use for authorization.

    Attributes
    ----------
    message_log : logging.Logger
        A log to which all REST request and response messages will be sent.  Attach a handler using
        `add_logger()` to capture these messages.

    filters : list of callable
        A collection of functions that will be called with each request and response object *prior* to logging the
        messages, allowing any sensitive information to be removed first.

    """

    PROFILE_PATH = '~/.sas/viya-api-profiles.yaml'

    def __init__(
        self,
        hostname,
        username=None,
        password=None,
        authinfo=None,
        protocol=None,
        port=None,
        verify_ssl=None,
        token=None,
    ):
        super(Session, self).__init__()

        # Determine whether or not server SSL certificates should be verified.
        if verify_ssl is None:
            verify_ssl = os.environ.get('SSLREQCERT', 'yes')
            verify_ssl = str(verify_ssl).lower() not in ('no', 'false')

        self._id = uuid4().hex
        self.message_log = logger.getChild('session.%s' % self._id)

        # If certificate path has already been set for SWAT package, make
        # Requests module reuse it.
        for k in ['SSLCALISTLOC', 'CAS_CLIENT_SSL_CA_LIST']:
            if k in os.environ:
                os.environ['REQUESTS_CA_BUNDLE'] = os.environ[k]
                break

        # If certificate path hasn't been specified in either environment
        # variable, replace the default adapter with one that will use the
        # machine's default SSL _settings.
        if 'REQUESTS_CA_BUNDLE' not in os.environ:
            if verify_ssl:
                # Skip hostname verification if IP address specified instead
                # of DNS name.  Prevents error from urllib3.
                try:
                    from urllib3.util.ssl_ import is_ipaddress
                except ImportError:
                    # is_ipaddres not present in older versions of urllib3
                    def is_ipaddress(hst):
                        return re.match(r"^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$", hst)

                verify_hostname = not is_ipaddress(hostname)
                adapter = SSLContextAdapter(assert_hostname=verify_hostname)

                self.mount('https://', adapter)

        # If we're skipping SSL verification, urllib3 will raise InsecureRequestWarnings on
        # every request.  Insert a warning filter so these warnings only appear on the first request.
        if not verify_ssl:
            from urllib3.exceptions import InsecureRequestWarning

            warnings.simplefilter('default', InsecureRequestWarning)

        self.filters = DEFAULT_FILTERS

        # Reuse an existing CAS connection if possible
        if swat and isinstance(hostname, swat.CAS):
            if isinstance(
                hostname._sw_connection, swat.cas.rest.connection.REST_CASConnection
            ):
                import base64

                # Use the httpAddress action to retieve information about
                # REST endpoints
                httpAddress = hostname.get_action('builtins.httpAddress')
                address = httpAddress()
                domain = address.virtualHost
                # httpAddress action may return virtualHost = ''
                # if this happens, try the CAS host
                if not domain:
                    domain = hostname._sw_connection._current_hostname
                protocol = address.protocol
                port = address.port
                auth = hostname._sw_connection._auth.decode('utf-8').replace(
                    'Basic ', ''
                )
                username, password = base64.b64decode(auth).decode('utf-8').split(':')
            else:
                raise ValueError(
                    "A 'swat.CAS' session can only be reused "
                    "when it's connected via the REST APIs."
                )
        else:
            url = urlsplit(hostname)

            # Extract http/https from domain name if present and protocol not explicitly given
            protocol = protocol or url.scheme
            domain = url.hostname or str(hostname)

        self._settings = {
            'protocol': protocol or 'https',
            'domain': domain,
            'port': port,
            'username': username,
            'password': password,
        }

        if self._settings['password'] is None:
            # Try to get credentials from .authinfo or .netrc files.
            # If no file path was specified, the default locations will
            # be checked.
            if 'swat' in sys.modules:
                auth = swat.utils.authinfo.query_authinfo(
                    domain, user=username, path=authinfo
                )
                self._settings['username'] = auth.get('user')
                self._settings['password'] = auth.get('password')

            # Not able to load credentials using SWAT.  Try Netrc.
            # TODO: IF a username was specified, verify that the credentials
            #       found are for that username.
            if self._settings['password'] is None:
                try:
                    parser = netrc.netrc(authinfo)
                    values = parser.authenticators(domain)
                    if values:
                        (
                            self._settings['username'],
                            _,
                            self._settings['password'],
                        ) = values
                except (OSError, IOError):
                    pass  # netrc throws if $HOME is not set

        # Set this prior authentication attempts
        self.verify = verify_ssl

        # Find a suitable authentication mechanism and build an auth header
        self.auth = self.get_auth(
            self._settings['username'], self._settings['password'], token
        )

        # Used for context manager
        self._old_session = current_session()
        current_session(self)

    def add_logger(self, handler, level=None):
        """Log session requests and responses.

        Parameters
        ----------
        handler : logging.Handler
            A Handler instance to use for logging the requests and responses.
        level : int, optional
            The logging level to assign to the handler.  Ignored if handler's
            logging level is already set.  Defaults to DEBUG.

        Returns
        -------
        handler


        .. versionadded:: 1.2.0

        """
        level = level or logging.DEBUG

        if handler.level == logging.NOTSET:
            handler.setLevel(level)

        self.message_log.addHandler(handler)

        if self.message_log.level == logging.NOTSET:
            self.message_log.setLevel(handler.level)

        return handler

    def add_stderr_logger(self, level=None):
        """Log session requests and responses to stderr.

        Parameters
        ----------
        level : int, optional
            The logging level of the handler.  Defaults to logging.DEBUG

        Returns
        -------
        logging.StreamHandler

        """
        return self.add_logger(logging.StreamHandler(), level)

    @versionadded(version='1.5.4')
    def as_swat(self, server=None, **kwargs):
        """Create a SWAT connection to a CAS server.

        Uses the authentication information from the session to establish a CAS connection using SWAT.

        Parameters
        ----------
        server : str, optional
            The logical name of the CAS server, not the hostname.  Defaults to "cas-shared-default".
        kwargs : any
            Additional arguments to pass to the `swat.CAS` constructor.  Can be used to override this method's
            default behavior or customize the CAS session.

        Returns
        -------
        swat.CAS
            An active SWAT connection

        Raises
        ------
        RuntimeError
            If `swat` package is not available.

        Examples
        --------
        >>> sess = Session('example.sas.com')
        >>> with sess.as_swat() as conn:
        ...    conn.listnodes()
        CASResults([('nodelist', Node List
                      name        role connected   IP Address
        0  example.sas.com  controller       Yes  127.0.0.1)])

        """
        server = server or 'cas-shared-default'

        if swat is None:
            raise RuntimeError(
                "The 'swat' package must be installed to create a SWAT connection."
            )

        # Construct the CAS server's URL
        url = '{}://{}/{}-http/'.format(
            self._settings['protocol'], self.hostname, server
        )

        kwargs.setdefault('hostname', url)

        # Starting in SWAT v1.8 oauth tokens could be passed directly in the password param.
        # Otherwise, use the username & password to re-authenticate.
        if version.parse(swat.__version__) >= version.parse('1.8'):
            kwargs['username'] = None
            kwargs['password'] = self.auth.access_token
        else:
            # Use this sessions info to connect to CAS unless user has explicitly give a value (even if None)
            kwargs.setdefault('username', self.username)
            kwargs.setdefault('password', self._settings['password'])

        orig_sslreqcert = os.environ.get('SSLREQCERT')

        # If SSL connections to microservices are not being verified, don't attempt
        # to verify connections to CAS - most likely certs are not in place.
        if not self.verify:
            os.environ['SSLREQCERT'] = 'no'

        try:
            cas = swat.CAS(**kwargs)
            cas.setsessopt(messagelevel='warning')
        finally:
            # Reset environment variable to whatever it's original value was
            if orig_sslreqcert:
                os.environ['SSLREQCERT'] = orig_sslreqcert

        return cas

    @property
    def username(self):
        return self._settings.get('username')

    @property
    def hostname(self):
        return self._settings.get('domain')

    def send(self, request, **kwargs):
        if self.message_log.isEnabledFor(logging.DEBUG):
            r = copy.deepcopy(request)
            for filter in self.filters:
                r = filter(r)

            self.message_log.debug(
                'HTTP/1.1 {verb} {url}\n{headers}\nBody:\n{body}'.format(
                    verb=r.method,
                    url=r.url,
                    headers='\n'.join(
                        '{}: {}'.format(k, v) for k, v in r.headers.items()
                    ),
                    body=_pformat(r.body),
                )
            )
        else:
            self.message_log.info('HTTP/1.1 %s %s', request.method, request.url)

        response = super(Session, self).send(request, **kwargs)

        if self.message_log.isEnabledFor(logging.DEBUG):
            r = copy.deepcopy(response)
            for filter in self.filters:
                r = filter(r)

            self.message_log.debug(
                'HTTP {status} {url}\n{headers}\nBody:\n{body}'.format(
                    status=r.status_code,
                    url=r.url,
                    headers='\n'.join(
                        '{}: {}'.format(k, v) for k, v in r.headers.items()
                    ),
                    body=_pformat(r.text),
                )
            )
        else:
            self.message_log.info('HTTP/1.1 %s %s', response.status_code, response.url)

        return response

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
    ):

        url = self._build_url(url)
        verify = verify or self.verify

        try:
            r = super(Session, self).request(
                method,
                url,
                params,
                data,
                headers,
                cookies,
                files,
                auth,
                timeout,
                allow_redirects,
                proxies,
                hooks,
                stream,
                verify,
                cert,
                json,
            )

            if r.status_code == 401:
                auth_header = r.headers.get('WWW-Authenticate', '').lower()

                # Access token expired, need to refresh it (if we can)
                if 'access token expired' in auth_header:
                    try:
                        self.auth = self.get_oauth_token(
                            refresh_token=self.auth.refresh_token
                        )

                        # Repeat the request
                        r = super(Session, self).request(
                            method,
                            url,
                            params,
                            data,
                            headers,
                            cookies,
                            files,
                            auth,
                            timeout,
                            allow_redirects,
                            proxies,
                            hooks,
                            stream,
                            verify,
                            cert,
                            json,
                        )
                    except exceptions.AuthorizationError:
                        pass

            return r
        except requests.exceptions.SSLError as e:
            if 'REQUESTS_CA_BUNDLE' not in os.environ:
                raise RuntimeError(
                    "SSL handshake failed.  The 'REQUESTS_CA_BUNDLE' "
                    "environment variable should contain the path to the CA "
                    "certificate.  Alternatively, set verify_ssl=False to "
                    "disable certificate verification."
                )
            raise e

    def get(self, url, **kwargs):
        return self.request('GET', url, **kwargs)

    def post(self, url, **kwargs):
        return self.request('POST', url, **kwargs)

    def put(self, url, **kwargs):
        return self.request('PUT', url, **kwargs)

    def head(self, url, **kwargs):
        return self.request('HEAD', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request('DELETE', url, **kwargs)

    def get_auth(self, username=None, password=None, token=None):
        """Attempt to authenticate with the Viya environment.

        If `username` and `password` or `token` are provided, they will be used exclusively.  If neither of these
        is specified, Kerberos authorization will be attempted.  If this fails, authorization using an
        OAuth2 authorization code will be attempted.  Be aware that this requires prompting the user for input and
        should NOT be used in a non-interactive session.

        Parameters
        ----------
        username : str, optional
            Username to use for authentication.
        password : str, optional
            Password to use for authentication.  Required if `username` is provided.
        token : str, optional
            An existing OAuth2 access token to use.

        Returns
        -------
        OAuth2Token
            Authorization header to use with requests to the evironent

        """

        # If explicit username & password were provided, use them.
        if username is not None and password is not None:
            return self.get_oauth_token(username, password)

        # If an existing access token was provided, use it
        if token is not None:
            return OAuth2Token(token)

        # Kerberos doesn't require any interruption to the user, so try that first if username/password
        # were not provided.
        try:
            token = self._get_token_with_kerberos()
            return OAuth2Token(token)
        except:
            logger.exception('Failed to connect with Kerberos')

        # Try loading cached credentials
        token = self.read_cached_token(self.PROFILE_PATH)

        if token is not None:
            return token

        # If we got this far, then no password and no kerberos.  Try prompting the user for an authorization code.
        auth_code = self.prompt_for_auth_code()
        return self.get_oauth_token(auth_code=auth_code)

    def get_oauth_token(
        self,
        username=None,
        password=None,
        auth_code=None,
        refresh_token=None,
        client_id=None,
        client_secret=None,
    ):
        """Request an OAuth2 access token using either a username & password or an auth token.

        Parameters
        ----------
        username : str, optional
            Username to use for authentication.
        password : str, optional
            Password to use for authentication.  Required if `username` is provided.
        auth_code : str, optional
            Authorization code to use.
        refresh_token : str, optinoal
            Valid refresh token to use.
        client_id : str, optional
            Client ID requesting access.  Use if connection to Viya should be made using a non-default client id.
        client_secret : str, optional
            Client secret for client requesting access.  Required if `client_id` is provided

        Returns
        -------
        OAuth2Token

        See Also
        --------
        :meth:`Session.prompt_for_auth_code`

        """

        if username is not None:
            data = {
                'grant_type': 'password',
                'username': username,
                'password': password,
            }
        elif auth_code is not None:
            data = {'grant_type': 'authorization_code', 'code': auth_code}
        elif refresh_token is not None:
            data = {'grant_type': 'refresh_token', 'refresh_token': refresh_token}
        else:
            raise ValueError(
                'Either username and password or auth_code parameters must be specified.'
            )

        client_id = client_id or os.environ.get('SASCTL_CLIENT_ID', 'sas.ec')
        client_secret = client_secret or os.environ.get('SASCTL_CLIENT_SECRET', '')

        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        url = self._build_url('/SASLogon/oauth/token')
        r = super(Session, self).post(
            url,
            headers=headers,
            data=data,
            auth=(client_id, client_secret),
            verify=self.verify,
        )

        # Raise user-friendly error messages if issue is known
        if r.status_code == 400 and auth_code is not None:
            j = r.json()
            if "Invalid authorization code" in j.get('error_description', ''):
                raise exceptions.AuthorizationError(
                    "Invalid authorization code: '%s'" % auth_code
                )
        if r.status_code == 401:
            if (
                r.json().get('error_description', '').lower() == 'bad credentials'
                and username is None
            ):
                raise exceptions.AuthenticationError(msg='Invalid client id or secret.')
            if r.json().get('error', '') == 'invalid_token':
                raise exceptions.AuthorizationError(
                    'Refresh token is incorrect, expired, or revoked.'
                )
            if username is not None:
                raise exceptions.AuthenticationError(username)

        r.raise_for_status()

        token = OAuth2Token(**r.json())

        # No reason to cache token if username & password were provided - they'd just be re-provided on future
        # connections.  Cache for auth code & refresh token to prevent frequent interruptions for the user.
        if username is None:
            self.cache_token(token, self.PROFILE_PATH)

        return token

    def prompt_for_auth_code(self, client_id=None):
        """Prompt the user open a URL to generate an auth code.

        Note that this halts program execution until input is received and should only be used for interactive sessions.

        Parameters
        ----------
        client_id : str, optional

        Returns
        -------
        str
            Authorization code that can be used to acquire an OAuth2 access token.

        See Also
        --------
        :meth:`Session.get_oauth_token`

        """
        client_id = client_id or os.environ.get('SASCTL_CLIENT_ID', 'sas.ec')

        # User must open this URL in a browser and then enter the auth code that's generated.
        url = (
            self._build_url('/SASLogon/oauth/authorize')
            + '?response_type=code&client_id='
            + client_id
        )
        message = (
            'Please use a web browser to login at the following URL to get your authorization code:\n'
            + url
        )
        print(message)
        auth_code = input('Authorization Code:')

        return auth_code

    def cache_token(self, token, path):
        """Write an OAuth2 token to the cache.

        Parameters
        ----------
        token : OAuth2Token
            Token to be cached.
        path : str
            Path to file containing cached tokens.

        Returns
        -------
        None

        """
        profiles = Session._read_token_cache(path)
        base_url = self._build_url('')

        # Top-level structure if no existing file was found
        if profiles is None:
            profiles = {'profiles': []}

        # Token values to be cached
        token = {
            'accesstoken': token.access_token,
            'refreshtoken': token.refresh_token,
            'tokentype': 'bearer',
            'expiry': token.expiration,
        }

        # See if there's an existing profile to update
        matches = [
            (i, p)
            for i, p in enumerate(profiles['profiles'])
            if p['baseurl'] == base_url
        ]
        if matches:
            idx, match = matches[0]
            match['oauthtoken'] = token
            profiles['profiles'][idx] = match
        else:
            profiles['profiles'].append(
                {'baseurl': base_url, 'name': None, 'oauthtoken': token}
            )

        Session._write_token_cache(profiles, path)

    def read_cached_token(self, path):
        """Read any cached access tokens from disk

        Parameters
        ----------
        path : str
            Path to file containing cached tokens.

        Returns
        -------
        OAuth2Token or None

        """
        # Map from field names in YAML to fields returned by SASLogon
        field_mappings = {
            'accesstoken': 'access_token',
            'refreshtoken': 'refresh_token',
            'expiry': 'expiration',
        }

        profiles = Session._read_token_cache(path)
        url = self._build_url('')

        # Couldn't read any profiles
        if profiles is None:
            return

        # Check each profile for a hostname match and return token if found
        for profile in profiles.get('profiles', []):
            baseurl = profile.get('baseurl', '').lower()

            # Return the cached token
            if baseurl == url.lower():
                data = profile.get('oauthtoken', {})
                token = {
                    field_mappings[k]: v for k, v in data.items() if k in field_mappings
                }
                token = OAuth2Token(**token)

                # Attempt to refresh if cached token has expired
                if token.is_expired:
                    try:
                        token = self.get_oauth_token(refresh_token=token.refresh_token)
                    except (
                        exceptions.AuthorizationError,
                        requests.exceptions.HTTPError,
                    ):
                        return

                # If refresh fails, dont return token, allow user to be prompted for login
                if not token.is_expired:
                    return token

    def _get_token_with_kerberos(self):
        """Authenticate with a Kerberos ticket."""
        if kerberos is None:
            raise RuntimeError(
                "Kerberos package not found.  Run 'pip "
                "install sasctl[kerberos]' to install."
            )

        user = self.username
        client_id = 'sas.tkmtrb'
        flags = kerberos.GSS_C_MUTUAL_FLAG | kerberos.GSS_C_SEQUENCE_FLAG
        service = 'HTTP@%s' % self._settings['domain']

        logger.info('Attempting Kerberos authentication to %s', service)

        url = self._build_url(
            '/SASLogon/oauth/authorize?client_id=%s&response_type=token' % client_id
        )

        # Get Kerberos challenge
        r = self.get(url, allow_redirects=False, verify=self.verify)

        if r.status_code != 401:
            raise ValueError(
                'Kerberos challenge response not received.  '
                'Expected HTTP 401 but received %s' % r.status_code
            )

        if 'www-authenticate' not in r.headers:
            raise ValueError(
                "Kerberos challenge response not received.  "
                "'WWW-Authenticate' header not received."
            )

        if 'Negotiate' not in r.headers['www-authenticate']:
            raise ValueError(
                "Kerberos challenge response not received.  "
                "'WWW-Authenticate' header contained '%s', "
                "expected 'Negotiate'." % r.headers['www-authenticate']
            )

        # Initialize a request to KDC for a ticket to access the service.
        _, context = kerberos.authGSSClientInit(service, principal=user, gssflags=flags)

        # Send the request.
        # NOTE: empty-string parameter required for initial call.
        kerberos.authGSSClientStep(context, '')

        # Get the KDC response
        auth_header = 'Negotiate %s' % kerberos.authGSSClientResponse(context)

        # Get the user that was used for authentication
        username = kerberos.authGSSClientUserName(context)

        # Drop @REALM from username and store
        if username is not None:
            self._settings['username'] = username.rsplit('@', maxsplit=1)[0]

        # Response to Kerberos challenge with ticket
        r = self.get(
            url,
            headers={'Authorization': auth_header},
            allow_redirects=False,
            verify=self.verify,
        )

        if 'Location' not in r.headers:
            raise ValueError(
                "Invalid authentication response." "'Location' header not received."
            )

        match = re.search('(?<=access_token=)[^&]*', r.headers['Location'])

        if match is None:
            raise ValueError(
                "Invalid authentication response.  'Location' "
                "header does not contain an access token."
            )

        return match.group(0)

    def _build_url(self, url):
        """Build a complete URL from a path by substituting in session parameters."""
        components = urlsplit(url)

        domain = components.netloc or self._settings['domain']

        # Add port if a non-standard port was specified
        if self._settings['port'] is not None and ':' not in domain:
            domain = '{}:{}'.format(domain, self._settings['port'])

        return urlunsplit(
            [
                components.scheme or self._settings['protocol'],
                domain,
                components.path,
                components.query,
                components.fragment,
            ]
        )

    @staticmethod
    def _read_token_cache(path):
        """Read cached OAuth2 access tokens from disk.

        Parameters
        ----------
        path : str
            Path to file containing cached tokens.

        Returns
        -------
        dict or None

        Raises
        ------
        RuntimeError
            If file permissions are too permissive.

        """
        yaml_file = os.path.expanduser(path)

        # See if a token has been cached for the hostname
        if os.path.exists(yaml_file):
            # Get bit flags indicating access permissions
            mode = os.stat(yaml_file).st_mode
            flags = oct(mode)[-3:]

            if flags != '600':
                raise RuntimeError(
                    'Unable to read profile cache.  '
                    'The file permissions for %s must be configured so that only the file owner has '
                    'read/write permissions (equivalent to 600 on Linux systems).'
                    % yaml_file
                )

            with open(yaml_file) as f:
                return yaml.load(f, Loader=yaml.FullLoader)

    @staticmethod
    def _write_token_cache(profiles, path):
        """

        Parameters
        ----------
        profiles : dict
        path : str

        Returns
        -------
        None

        """
        yaml_file = os.path.expanduser(path)

        # Create parent .sas folder if needed
        sas_dir = os.path.dirname(yaml_file)
        if not os.path.exists(sas_dir):
            os.mkdir(sas_dir)

        with open(yaml_file, 'w') as f:
            yaml.dump(profiles, f)

        # Get bit flags indicating access permissions
        mode = os.stat(yaml_file).st_mode
        flags = oct(mode)[-3:]

        # Ensure access to file is restricted
        if flags != '600':
            os.chmod(yaml_file, 0o600)

    def __enter__(self):
        super(Session, self).__enter__()

        # Make this the current session
        # self._old_session = current_session()
        # current_session(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore previous current session
        current_session(self._old_session)

        super(Session, self).__exit__()

    def __str__(self):
        return (
            "{class_}(hostname='{hostname}', username='{username}', "
            "protocol='{protocol}', verify_ssl={verify})".format(
                class_=type(self).__name__,
                hostname=self.hostname,
                username=self.username,
                protocol=self._settings.get('protocol'),
                verify=self.verify,
            )
        )


class PageIterator:
    """Iterates through a collection that must be "paged" from the server.

    Pages contain a batch of items from the overall collection.  Iterates the
    series of pages and returns a single batch of items each time `next()` is
    called.

    Parameters
    ----------
    obj : RestObj
        An instance of `RestObj` containing any initial items and a link to
        retrieve additional items.
    session : Session
        The `Session` instance to use for requesting additional items.  Defaults
        to current_session()
    threads : int
        Number of threads allocated to downloading additional pages.

    Yields
    ------
    List[RestObj]
        Items contained in the current page

    """

    def __init__(self, obj, session=None, threads=4):
        self._num_threads = threads

        # Session to use when requesting items
        self._session = session or current_session()
        self._pool = None
        self._requested = []

        link = get_link(obj, 'next')

        # Dissect the "next" link so it can be reformatted and used by
        # parallel threads
        if link is None:
            self._min_queue_len = 0
            self._start = 0
            self._limit = 0
        if link is not None:
            link = link['href']
            start = re.search(r'(?<=start=)[\d]+', link)
            limit = re.search(r'(?<=limit=)[\d]+', link)

            # Construct a new link with format params
            # Result is "/spam/spam?start={start}&limit={limit}"
            link = (
                link[: start.start()]
                + '{start}'
                + link[start.end() : limit.start()]
                + '{limit}'
                + link[limit.end() :]
            )

            self._start = int(start.group())
            self._limit = int(limit.group())

            # Length at which to beging requesting new items from the server
            self._min_queue_len = self._limit

        self._next_link = link

        # Store the current items to iterate over
        self._obj = obj

    def __next__(self):
        if self._pool is None:
            self._pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=self._num_threads
            )

        # Send request for new pages if we don't have enough cached
        num_req_needed = self._num_threads - len(self._requested)
        for _ in range(num_req_needed):
            self._requested.append(self._pool.submit(self._request_async, self._start))
            self._start += self._limit

        # If this is the first time next() has been called, return the items
        # contained in the initial page.
        if self._obj is not None:
            result = [RestObj(x) for x in self._obj['items']]
            self._obj = None
            return result

        # Make sure the next page has been received
        if self._requested:
            items = self._requested.pop(0).result()

            if not items:
                raise StopIteration

            return items

        raise StopIteration

    def __iter__(self):
        # All Iterators are also Iterables
        return self

    def _request_async(self, start):
        """Used by worker threads to retrieve next batch of items."""

        if self._next_link is None:
            return []

        # Format the link to retrieve desired batch
        link = self._next_link.format(start=start, limit=self._limit)
        r = get(link, format='json', session=self._session)
        r = RestObj(r)
        return [RestObj(x) for x in r['items']]


class PagedItemIterator:
    """Iterates through a collection that must be "paged" from the server.

    Uses `PageIterator` to transparently download pages of items from the server
    as needed.

    Parameters
    ----------
    obj : RestObj
        An instance of `RestObj` containing any initial items and a link to
        retrieve additional items.
    session : Session
        The `Session` instance to use for requesting additional items.  Defaults
        to current_session()
    threads : int
        Number of threads allocated to downloading additional items.

    Yields
    ------
    RestObj

    Notes
    -----
    Value returned by len() is an approximate count of the items available.  The actual
    number of items returned may be greater than or less than this number.

    See Also
    --------
    PageIterator

    """

    def __init__(self, obj, session=None, threads=4):
        # Iterates over whole pages of items
        self._pager = PageIterator(obj, session, threads)

        # Store items from latest page that haven't been returned yet.
        self._cache = []

        # Total number of items to iterate over
        if 'count' in obj:
            # NOTE: "count" may be an (over) estimate of the number of items available
            #       since some may be inaccessible due to user permissions & won't actually be returned.
            self._count = int(obj.count)
        else:
            self._count = len(obj['items'])

    def __len__(self):
        return self._count

    def __next__(self):
        # Get next page of items if we're currently out
        if not self._cache:
            self._cache = next(self._pager)

        # Return the next item
        if self._cache:
            self._count -= 1
            return self._cache.pop(0)

        raise StopIteration()

    def __iter__(self):
        return self


class PagedListIterator:
    """Iterates over an instance of PagedList

    Parameters
    ----------
    l : list-like

    """

    def __init__(self, l):
        self.__list = l
        self.__index = 0

    def __next__(self):
        if self.__index >= len(self.__list):
            raise StopIteration

        try:
            item = self.__list[self.__index]
            self.__index += 1
            return item
        except IndexError:
            # Because PagedList length is approximate, iterating can result in
            # indexing outside the array.  Just stop the iteration if that occurs.
            raise StopIteration

    def __iter__(self):
        return self


class PagedList(list):
    """List that dynamically loads items from the server.

    Parameters
    ----------
    obj : RestObj
        An instance of `RestObj` containing any initial items and a link to
        retrieve additional items.
    session : Session, optional
        The `Session` instance to use for requesting additional items.  Defaults
        to current_session()
    threads : int, optional
        Number of threads allocated to loading additional items.

    Notes
    -----
    Value returned by len() is an approximate count of the items available.  The actual
    length is not known until all items have been pulled from the server.

    See Also
    --------
    PagedItemIterator

    """

    def __init__(self, obj, session=None, threads=4):
        super(PagedList, self).__init__()
        self._paged_items = PagedItemIterator(obj, session=session, threads=threads)

        # Go ahead and add the items that were initially returned.
        # Do this by "paging" so iterator remains at the correct spot.
        for _ in range(len(obj['items'])):
            self.append(next(self._paged_items))

        # Assume that server has more items available
        self._has_more = True

    def __len__(self):
        if self._has_more:
            # Estimate the total length as items downloaded + items still on server
            return super(PagedList, self).__len__() + len(self._paged_items)
        else:
            # We've pulled everything from the server, so we have an exact length now.
            return super(PagedList, self).__len__()

    def __iter__(self):
        return PagedListIterator(self)

    def __getslice__(self, i, j):
        # Removed from Py3.x but still implemented in CPython built-in list
        # Override to ensure __getitem__ is used instead.
        return self.__getitem__(slice(i, j))

    def __getitem__(self, item):
        if hasattr(item, 'stop'):
            # `item` is a slice
            # if no stop was specified, assume full length
            idx = item.stop or len(self)
        else:
            idx = int(item)

            # Support negative indexing.  Need to load items up to len() - idx.
            if idx < 0:
                idx = len(self) + idx

        try:
            # Iterate through server-side pages until we've loaded
            # the item at the requested index.
            while super(PagedList, self).__len__() <= idx:
                n = next(self._paged_items)
                self.append(n)

        except StopIteration:
            # We've hit the end of the paging so the server has no more items to retrieve.
            self._has_more = False

        # Get the item from the list
        return super(PagedList, self).__getitem__(item)

    def __repr__(self):
        string = super(PagedList, self).__repr__()

        # If the list has more "items" than are stored in the underlying list
        # then there are more downloads to make.
        if len(self) - super(PagedList, self).__len__() > 0:
            string = string.rstrip(']') + ', ...]'

        return string


def is_uuid(id_):
    try:
        UUID(str(id_))
        return True
    except (ValueError, TypeError):
        return False


def get(path, **kwargs):
    """Send a GET request.

    Parameters
    ----------
    path : str
        The path portion of the URL.
    kwargs : any
        Passed to `request`.

    Returns
    -------
    RestObj or None
        The results or None if the resource was not found.

    """
    try:
        return request('get', path, **kwargs)
    except HTTPError as e:
        if e.code == 404:
            return None  # Resource not found
        raise e


def head(path, **kwargs):
    """Send a HEAD request.

    Parameters
    ----------
    path : str
        The path portion of the URL.
    kwargs : any
        Passed to `request`.

    Returns
    -------
    RestObj

    """
    return request('head', path, **kwargs)


def post(path, **kwargs):
    """Send a POST request.

    Parameters
    ----------
    path : str
        The path portion of the URL.
    kwargs : any
        Passed to `request`.

    Returns
    -------
    RestObj

    """
    return request('post', path, **kwargs)


def put(path, item=None, **kwargs):
    """Send a PUT request.

    Parameters
    ----------
    path : str
        The path portion of the URL.
    item : RestObj, optional
        A existing object to PUT.  If provided, ETag and Content-Type headers
        will automatically be specified.
    kwargs : any
        Passed to `request`.

    Returns
    -------
    RestObj

    """
    # If call is in the format put(url, RestObj), automatically fill in header
    # information
    if item is not None and isinstance(item, RestObj):
        get_headers = getattr(item, '_headers', None)
        if get_headers is not None:
            # Update the headers param if it was specified
            headers = kwargs.pop('headers', {})
            headers.setdefault('If-Match', get_headers.get('etag'))
            headers.setdefault('Content-Type', get_headers.get('content-type'))
            return request('put', path, json=item, headers=headers)

    return request('put', path, **kwargs)


def delete(path, **kwargs):
    """Send a DELETE request.

    Parameters
    ----------
    path : str
        The path portion of the URL.
    kwargs : any
        Passed to `request`.

    Returns
    -------
    RestObj

    """
    return request('delete', path, **kwargs)


def request(verb, path, session=None, format='auto', **kwargs):
    """Send an HTTP request with a session.

    Parameters
    ----------
    verb : str
        A valid HTTP request verb.
    path : str
        Path portion of URL to request.
    session : Session, optional
        Defaults to `current_session()`.
    format : {'auto', 'rest', 'response', 'content', 'json', 'text'}
        The format of the return response.  Defaults to `auto`.
        rest: `RestObj` constructed from JSON.
        response: the raw `Response` object.
        content: Response.content
        json: Response.json()
        text: Response.text
        auto: `RestObj` constructed from JSON if possible, otherwise same as
              `text`.
    kwargs : any
        Additional arguments are passed to the session `request` method.

    Returns
    -------

    """
    session = session or current_session()

    if session is None:
        raise TypeError('No `Session` instance found.')

    format = 'auto' if format is None else str(format).lower()
    if format not in ('auto', 'response', 'content', 'text', 'json', 'rest'):
        raise ValueError

    response = session.request(verb, path, **kwargs)

    if 400 <= response.status_code <= 599:
        raise HTTPError(
            response.url, response.status_code, response.text, response.headers, None
        )

    # Return the raw response if requested
    if format == 'response':
        return response
    if format == 'json':
        return response.json()
    if format == 'text':
        return response.text
    if format == 'content':
        return response.content
    try:
        obj = _unwrap(response.json())

        # ETag is required to update any object
        # May not be returned on all responses (e.g. listing
        # multiple objects)
        if isinstance(obj, RestObj):
            obj._headers = response.headers
        return obj
    except ValueError:
        if format == 'rest':
            return RestObj()
        return response.text


def get_link(obj, rel):
    """Get link information from a resource.

    Parameters
    ----------
    obj : dict
    rel : str

    Returns
    -------
    dict

    """
    if isinstance(obj, dict) and 'links' in obj:
        if isinstance(obj['links'], dict):
            return obj['links'].get(rel)

        links = [l for l in obj.get('links', []) if l.get('rel') == rel]
        if not links:
            return None
        if len(links) == 1:
            return links[0]
        return links

    if isinstance(obj, dict) and 'rel' in obj and obj['rel'] == rel:
        # Object is already a link, just return it
        return obj


def request_link(obj, rel, **kwargs):
    """Request a link from a resource.

    Parameters
    ----------
    obj : dict
    rel : str
    kwargs : any
        Passed to :function:`request`

    Returns
    -------

    """
    link = get_link(obj, rel)

    if link is None:
        raise ValueError("Link '%s' not found in object %s." % (rel, obj))

    return request(link['method'], link['href'], **kwargs)


def uri_as_str(obj):
    """Get the URI of a resource in string format.

    Parameters
    ----------
    obj : str or dict
        Strings are assumed to be URIs and are returned as is.  Dictionary
        objects will be checked for a `self` URI and returned if found.

    Returns
    -------
    str
        Resource URI or None if not found

    """
    if isinstance(obj, dict):
        link = get_link(obj, 'self')
        if isinstance(link, dict):
            return link.get('uri')

    return obj


def _unwrap(json):
    """Converts a JSON response to one or more `RestObj` instances.

    If the JSON contains a .items property, only those items are converted and returned.

    Parameters
    ----------
    json

    Returns
    -------

    """
    if 'items' in json:
        if len(json['items']) == 1:
            return RestObj(json['items'][0])
        if len(json['items']) > 1:
            return PagedList(RestObj(json))
        return []

    return RestObj(json)


def _build_crud_funcs(path, single_term=None, plural_term=None, service_name=None):
    """Utility method for defining simple functions to perform CRUD operations on a REST endpoint.

    Parameters
    ----------
    path : str
        URL path to use for the requests
    single_term : str
        English name of the item being manipulated. Defaults to `plural_term`.
    plural_term : str
        English name of the items being manipulated. Defaults to the last segment of `path`.
    service_name : str
        Name of the service under which the command will be listed in the `sasctl` CLI.  Defaults to `plural_term`.

    Returns
    -------
    functions : tuple
        tuple of CRUD functions: list_items, get_item, update_item, delete_item


    Examples
    --------

    >>> list_spam, get_spam, update_spam, delete_spam = _build_crud_funcs('/spam')

    """

    @sasctl_command('list')
    def list_items(filter=None):
        """List all {items} available in the environment.

        Parameters
        ----------
        filter : str, optional

        Returns
        -------
        list
            A list of dictionaries containing the {items}.

        Notes
        -----
        See the filtering_ reference for details on the `filter` parameter.

        .. _filtering: https://developer.sas.com/reference/filtering/

        """
        params = 'filter={}'.format(filter) if filter is not None else {}

        results = get(path, params=params)
        return results if isinstance(results, list) else [results]

    @sasctl_command('get')
    def get_item(item, refresh=False):
        """Returns a {item} instance.

        Parameters
        ----------
        item : str or dict
            Name, ID, or dictionary representation of the {item}.
        refresh : bool, optional
            Obtain an updated copy of the {item}.

        Returns
        -------
        RestObj or None
            A dictionary containing the {item} attributes or None.

        Notes
        -------
        If `item` is a complete representation of the {item} it will be returned unless `refresh` is set.  This
        prevents unnecessary REST calls when data is already available on the client.

        """

        # If the input already appears to be the requested object just return it, unless
        # a refresh of the data was explicitly requested.
        if isinstance(item, dict) and all(k in item for k in ('id', 'name')):
            if refresh:
                item = item['id']
            else:
                return item

        if is_uuid(item):
            return get(path + '/{id}'.format(id=item))
        results = list_items(filter='eq(name, "{}")'.format(item))

        # Not sure why, but as of 19w04 the filter doesn't seem to work.
        for result in results:
            if result['name'] == str(item):
                # Make a request for the specific object so that ETag is
                # included, allowing updates.
                if get_link(result, 'self'):
                    return request_link(result, 'self')

                id_ = result.get('id', result['name'])
                return get(path + '/{id}'.format(id=id_))

        return None

    @sasctl_command('update')
    def update_item(item):
        """Updates a {item} instance.

        Parameters
        ----------
        item : dict

        Returns
        -------
        None

        """

        headers = getattr(item, '_headers', None)
        if headers is None or headers.get('etag') is None:
            raise ValueError('Could not find ETag for update of %s.' % item)

        id_ = getattr(item, 'id', None)
        if id_ is None:
            raise ValueError('Could not find property `id` for update of %s.' % item)

        headers = {
            'If-Match': item._headers.get('etag'),
            'Content-Type': item._headers.get('content-type'),
        }

        return put(path + '/%s' % id_, json=item, headers=headers)

    @sasctl_command('delete')
    def delete_item(item):
        """Deletes a {item} instance.

        Parameters
        ----------
        item

        Returns
        -------
        None

        """

        # Try to find the item if the id can't be found
        if not (isinstance(item, dict) and 'id' in item):
            item = get_item(item)

        if isinstance(item, dict) and 'id' in item:
            item = item['id']

        if is_uuid(item):
            return delete(path + '/{id}'.format(id=item))
        raise ValueError("Unrecognized id '%s'" % item)

    # Pull object name from path if unspecified (many paths end in /folders or /repositories).
    plural_term = plural_term or str(path).split('/')[-1]
    single_term = single_term or plural_term
    service_name = service_name or plural_term
    service_name = service_name.replace(' ', '_')

    for func in [list_items, get_item, update_item, delete_item]:
        func.__doc__ = func.__doc__.format(item=single_term, items=plural_term)
        func._cli_service = service_name

        prefix = func.__name__.split('_')[0] + '_'
        suffix = plural_term if prefix == 'list_' else single_term
        func.__name__ = prefix + suffix

    return list_items, get_item, update_item, delete_item


def _build_is_available_func(service_root):
    def is_available():
        """Checks if the service is currently available.

        Returns
        -------
        bool

        """
        response = current_session().head(service_root + '/')
        return response.status_code == 200

    return is_available


@versionadded(version='1.5.6')
def platform_version():
    """Get the version of the SAS Viya platform to which sasctl is connected.

    Returns
    -------
    string : {'3.5', '4.0'}

        SAS Viya version number

    """
    from .services import model_repository as mr

    response = mr.info()
    buildVersion = response.get('build')['buildVersion']
    try:
        if buildVersion[0:4] == '3.7.':
            return '3.5'
        elif float(buildVersion[0:4]) >= 3.10:
            return '4.0'
    except ValueError:
        pass  # Version could not be found.  Return None instead.



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import argparse
import inspect
import json
import logging
import os
import pkgutil
import warnings
from collections import namedtuple, defaultdict
from importlib import import_module
from pprint import pprint


ArgInfo = namedtuple('ArgInfo', ['name', 'type', 'required', 'default', 'doc'])


def sasctl_command(name, subname=None):
    """Decorator that tags the function as being usable from the command line.

    Parameters
    ----------
    name : str
        the name of the command that will be shown on the command line.
    subname : str
        the name of the service that the command will be listed under

    Returns
    -------
    function

    Examples
    --------
    Define a command called 'cmd' not associated with a service
    >>> @sasctl_command('cmd')
    >>> def func():
            ...

    Define a command called 'cmd' associated with the 'svc' service
    >>> @sasctl_command('svc', 'cmd')
    >>> def func():
            ...

    Define a command and allow it's name and service to be auto-assigned
    >>> @sasctl_command
    >>> def func():
            ...

    """

    def decorator(func):
        if isinstance(name, str):
            if isinstance(subname, str):
                command_name = subname
                service_name = name
            else:
                command_name = name
                service_name = subname
        else:
            command_name = func.__name__
            if any(
                command_name.startswith(x)
                for x in ['list_', 'update_', 'get_', 'create_', 'delete_']
            ):
                parts = command_name.split('_')
                command_name = parts[0]
                service_name = parts[-1]
            else:
                service_name = subname

        def parse_args():
            """Retrieve argument metadata from function signature and docstring."""
            arg_spec = inspect.getargspec(func)
            defaults = list(arg_spec.defaults) if arg_spec.defaults is not None else []
            required = [True] * (len(arg_spec.args) - len(defaults)) + [False] * len(
                defaults
            )
            defaults = [None] * (len(arg_spec.args) - len(defaults)) + defaults
            types = []
            help_doc = []

            doc = inspect.getdoc(func)
            if doc and doc.find('Parameters\n'):
                doc_lines = doc[doc.find('Parameters\n') :].splitlines()
                doc_lines.pop(0)  # First line is "Parameters"

                if doc_lines and doc_lines[0].startswith('---'):
                    doc_lines.pop(
                        0
                    )  # Discard ----------- line under "Parameters" heading

                while doc_lines:
                    var = doc_lines.pop(0)

                    if var.startswith('Returns') or var.strip() == '':
                        break

                    if ':' in var:
                        types.append(var.split(':')[-1].strip())
                    else:
                        types.append('str')

                    if doc_lines and doc_lines[0].startswith('    '):
                        help_doc.append(doc_lines.pop(0).strip())
                    else:
                        help_doc.append('')
            else:
                types = ['str'] * len(arg_spec.args)
                help_doc = [None] * len(arg_spec.args)

            return [
                ArgInfo(n, t, r, d, o)
                for n, t, r, d, o in zip(
                    arg_spec.args, types, required, defaults, help_doc
                )
            ]

        func._cli_command = command_name
        func._cli_service = service_name
        func._cli_arguments = parse_args

        return func

    if callable(name):
        # allow direct decoration without arguments
        return decorator(name)
    return decorator


def _find_services(module='sasctl'):
    """Recursively find all functions in all modules that have been decorated as CLI commands."""
    m = __import__(module, fromlist=[''])  # returns a module

    def find_recurse(module, services):
        for obj in dir(module):
            obj = getattr(module, obj)
            if hasattr(obj, '_cli_command') and hasattr(obj, '_cli_service'):
                services[obj._cli_service][obj._cli_command] = obj

            elif type(obj).__module__.startswith('sasctl._services'):
                for atr in dir(obj):
                    atr = getattr(obj, atr)
                    if hasattr(atr, '_cli_command') and hasattr(atr, '_cli_service'):
                        services[atr._cli_service][atr._cli_command] = atr

        # recurse into submodules
        submodules = pkgutil.iter_modules(getattr(module, '__path__', []))

        for submodule in submodules:

            # ModuleInfo returned py 3.6 has .name
            # Tuple of (module_loader, name, ispkg) returned by older versions
            submodule_name = getattr(submodule, 'name', submodule[1])

            # TODO: Temporary until pzmm fully merged with sasctl
            if submodule_name == 'pzmm':
                continue

            submodule = import_module('.' + submodule_name, package=module.__name__)
            # if hasattr(submodule, 'name'):
            #     # ModuleInfo returned py 3.6
            #     submodule = import_module('.' + submodule.name, package=module.__name__)
            # else:
            #     # Tuple of (module_loader, name, ispkg) returned by older versions
            #     submodule = import_module('.' + submodule[1], package=module.__name__)
            services = find_recurse(submodule, services)

        return services

    services = find_recurse(m, defaultdict(dict))
    return services


def _get_func_description(func):
    description = getattr(func, '__doc__', '')

    lines = description.split('\n')

    if lines:
        return lines[0]


def _build_parser(services):
    from sasctl import __version__

    # TODO: Set command docstring

    # Create standard, top-level arguments
    parser = argparse.ArgumentParser(
        prog='sasctl', description='sasctl interacts with a SAS Viya environment.'
    )
    parser.add_argument(
        '-k', '--insecure', action='store_true', help='skip SSL verification'
    )
    parser.add_argument(
        '-f', '--format', choices=['json'], default='json', help='output format'
    )
    parser.add_argument('-v', '--verbose', action='count')
    parser.add_argument(
        '--version', action='version', version='%(prog)s ' + __version__
    )

    subparsers = parser.add_subparsers(title='service', dest='service')
    subparsers.required = True

    for service, commands in services.items():
        service_parser = subparsers.add_parser(service)
        service_subparser = service_parser.add_subparsers(
            title='command', dest='command'
        )
        service_subparser.required = True

        # Add the command and arguments for each command
        for command in commands:
            func = services[service][command]

            cmd_parser = service_subparser.add_parser(
                command, help=_get_func_description(func)
            )

            for arg in func._cli_arguments():
                if arg.name in ('self', 'cls'):
                    continue

                if arg.required:
                    cmd_parser.add_argument(arg.name, help=arg.doc)
                else:
                    cmd_parser.add_argument(
                        '--' + arg.name,
                        required=arg.required,
                        default=arg.default,
                        help=arg.doc,
                    )

    return parser


def main(args=None):
    """Main entry point when executed as a command line utility."""
    from sasctl import Session, current_session

    # Find all services and associated commands
    services = _find_services()

    parser = _build_parser(services)
    args = parser.parse_args(args)

    if args.verbose is None or args.verbose == 0:
        lvl = logging.WARNING
    elif args.verbose == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    handler = logging.StreamHandler()
    handler.setLevel(lvl)
    logging.getLogger('sasctl.core').addHandler(handler)
    logging.getLogger('sasctl.core').setLevel(lvl)

    warnings.simplefilter('ignore')

    func = services[args.service][args.command]
    kwargs = vars(args).copy()

    # Remove args that shouldn't be passed to the underlying command
    for k in ['command', 'service', 'insecure', 'verbose', 'format']:
        kwargs.pop(k, None)

    username = os.environ.get('SASCTL_USER_NAME')
    password = os.environ.get('SASCTL_PASSWORD')
    server = os.environ.get('SASCTL_SERVER_NAME')

    if server is None:
        parser.error(
            "Hostname must be specified in the 'SASCTL_SERVER_NAME' environment variable."
        )

    verify_ssl = not args.insecure

    try:
        #  current_session() should never be set when executing from the
        #  command line but it allows us to provide a pre-created session
        #  during testing
        with current_session() or Session(
            server, username, password, verify_ssl=verify_ssl
        ):
            result = func(**kwargs)
            if isinstance(result, list):
                pprint([str(x) for x in result])
            elif isinstance(result, dict) and args.format == 'json':
                print(json.dumps(result, indent=2))
            else:
                pprint(result)
    except RuntimeError as e:
        parser.error(e)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .service import Service
from ..core import RestObj

LOW = 0
MEDIUM = 1
HIGH = 2

_LOD_VALUES = {LOW: 'thumbnail', MEDIUM: 'normal', HIGH: 'entireSection'}


class ReportImages(Service):
    """Delivers SVG images representing elements of a report.

    The images are suitable to the current user, taking row-level-permissions
    and other factors into consideration.

    See Also
    --------
    `REST Documentation <https://developer.sas.com/apis/rest/Visualization/#
    report-images>`_

    """

    _SERVICE_ROOT = 'reportImages'

    LOD_LOW = LOW
    """Renders the content with minimal details for creating thumbnails."""

    LOD_MEDIUM = MEDIUM
    """Renders the primary components of the content."""

    LOD_HIGH = HIGH
    """Renders all of the components of content."""

    @classmethod
    def get_images_async(cls, report, section=None, elements=None, size=None, lod=None):
        """Create a job to render images of reports or report sections.

        Report contents are sized and arranged dynamically to accommodate
        browser viewing.  Some or all of these contents may be rendered at any
        time to produce an SVG image of the desired size.  This is done by
        creating a render job which will create the requested images and return
        their download links.

        Parameters
        ----------
        report : str or RestObj
            Report to retrieve images from specified as either a URI or a
            `RestObj` instance as returned from `reports.get_report`.
        section : int, optional
            A specific report section to retrieve, specified by a zero-based
            index.  If not specified, all sections will be returned as separate
            images.
        elements : List[Tuple[str, str]], optional
            Iterable of tuples containing the element name and the size of the
            element when rendered.
        size : str or Tuple[int, int], optional
            Size (in pixels) of the resulting images.  Ignored if `elements` is
            specified.
        lod : int, optional
            Level of detail of the rendered image(s).  Defaults to HIGH.

        Returns
        -------
        RestObj
            The render job

        """

        if isinstance(report, RestObj):
            report = cls.get_link(report, 'self')['uri']

        if isinstance(size, tuple):
            size = '%dx%d' % size

        if size is None:
            size = '640x480'

        lod = lod or HIGH

        if lod not in _LOD_VALUES.keys():
            raise ValueError(
                "LOD value of '%s' is invalid.  Expected one of " "LOW, MEDIUM, HIGH."
            )

        # LOD.HIGH will render the entire section
        # If request is to render individual elements, we must reduce LOD.
        if elements is not None and lod == HIGH:
            lod = MEDIUM

        # Convert integer enum to string value
        lod = _LOD_VALUES[lod]

        formatted_elements = []

        if elements is not None:
            selection = 'visualElements'

            # In the case where a single element was passed, wrap it in a list
            if isinstance(elements, (str, dict)):
                elements = [elements]

            # In the case where a single tuple was passed, wrap it in a list
            # Don't do this if a tuple of tuples was passed.
            if isinstance(elements, tuple) and not isinstance(elements[0], tuple):
                elements = [elements]

            for element in elements:
                if isinstance(element, tuple) and len(element) == 2:
                    # Get element & target size if both were provided
                    elem_name, elem_size = element
                else:
                    # Otherwise, assume top-level size value will be used.
                    elem_name = element
                    elem_size = size

                # If element was passed as a RestObj, extract the name
                if hasattr(elem_name, 'name'):
                    elem_name = elem_name.name

                # If size was passed as a tuple, format as string
                if isinstance(elem_size, tuple):
                    elem_size = '%dx%d' % elem_size

                formatted_elements.append(dict(name=elem_name, size=elem_size))

        elif section is not None:
            selection = 'report'
        else:
            selection = 'perSection'

        job = cls.post(
            '/jobs#requestBody',
            json={
                'reportUri': report,
                'layoutType': lod,
                'selectionType': selection,
                'size': size,
                'specificVisualElements': formatted_elements,
                'sectionIndex': section,
            },
        )

        return job

    @classmethod
    def get_images(cls, report, section=None, elements=None, size=None, lod=None):
        """Render images of reports or report sections.

        Report contents are sized and arranged dynamically to accommodate
        browser viewing.  Some or all of these contents may be rendered at any
        time to produce an SVG image of the desired size.  A rendering job will
        automatically be created and the resulting images returned upon job
        completion.

        Parameters
        ----------
        report : str or RestObj
            Report to retrieve images from specified as either a URI or a
            `RestObj` instance as returned from `reports.get_report`.
        section : int, optional
            A specific report section to retrieve, specified by a zero-based
            index.  If not specified, all sections will be returned as separate
            images.
        elements : List[Tuple[str, str]], optional
            Iterable of tuples containing the element name and the size of the
            element when rendered.
        size : str or Tuple[int, int], optional
            Size (in pixels) of the resulting images.  Ignored if `elements` is
            specified.
        lod : `LOD`, optional
            Level of detail of the rendered image(s).  Defaults to LOD.HIGH.

        Returns
        -------
        List[bytes]
            List of binary strings, each representing one SVG image.

        Examples
        --------
        Render a single element of a report

        >>> from sasctl.services import reports
        >>> report = reports.get_report('Parrot Status')
        >>> report_images.get_images(report, elements=('Norwegian Blue', '640x480'))
        [b'<svg>beautiful_plummage</svg>']

        Render each page of the report

        >>> from sasctl.services import reports
        >>> report = reports.get_report('Parrot Status')
        >>> report_images.get_images(report)
        [b'<svg>Customer Complaints</svg>', b'<svg>Deceased</svg>']

        Render the 2nd page of a report

        >>> from sasctl.services import reports
        >>> report = reports.get_report('Parrot Status')
        >>> report_images.get_images(report, section=1)
        [b'<svg>Deceased</svg>']

        """
        job = cls.get_images_async(
            report=report, section=section, elements=elements, size=size, lod=lod
        )

        result = cls._monitor_job(job)

        images = [
            cls.request_link(img, 'image', format='content') for img in result.images
        ]

        return images



# Copyright (c) 2020, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# %%
from pathlib import Path
import zipfile
import io

# %%

class ZipModel():
    
    @staticmethod
    def zipFiles(fileDir, modelPrefix):
        '''
        Combines all JSON files with the model pickle file and associated score code file
        into a single archive ZIP file.
        
        Parameters
        ---------------
        fileDir : string
            Location of *.json, *.pickle, and *Score.py files.
        modelPrefix : string
            Variable name for the model to be displayed in SAS Open Model Manager 
            (i.e. hmeqClassTree + [Score.py || .pickle]).
            
        Yields
        ---------------
        '*.zip'
            Archived ZIP file for importing into SAS Open Model Manager. In this form,
            the ZIP file can be imported into SAS Open Model Manager.
        '''
        
        fileNames = []
        fileNames.extend(sorted(Path(fileDir).glob('*.json')))
        fileNames.extend(sorted(Path(fileDir).glob('*Score.py')))
        fileNames.extend(sorted(Path(fileDir).glob('*.pickle')))
        # Include H2O.ai MOJO files 
        fileNames.extend(sorted(Path(fileDir).glob('*.mojo')))
        
        with zipfile.ZipFile(Path(fileDir) / (modelPrefix + '.zip'), mode='w') as zFile:
            for file in fileNames:
                zFile.write(file, arcname=file.name)
                
        with open(Path(fileDir) / (modelPrefix + '.zip'), 'rb') as zipFile:
            return io.BytesIO(zipFile.read())



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import functools
import textwrap
import warnings


class ExperimentalWarning(UserWarning):
    """Warning raised by @experimental decorator."""

    pass


def _insert_docstring_text(func, text):
    docstring = func.__doc__ or ''

    # Dedent the existing docstring.  Multi-line docstrings are only indented
    # after the first line, so split before dedenting.
    if '\n' in docstring:
        first_line, remainder = docstring.split('\n', 1)
        docstring = first_line + '\n' + textwrap.dedent(remainder)
    else:
        docstring = textwrap.dedent(docstring)

    docstring = docstring.strip()

    # Sphinx formatting requires 2 blank lines after a numpydoc section or one
    # blank line after another Sphinx directive.
    if docstring.split('\n')[-1].startswith('.. '):
        # Last line is another Sphinx directive
        gap = '\n\n'
    else:
        gap = '\n\n\n'

    return docstring + gap + text + '\n'


def deprecated(reason=None, version=None, removed_in=None):
    """Decorate a function or class to designated it as deprecated.

    Will raise a `DeprecationWarning` when used and automatically adds a
    Sphinx '.. deprecated::' directive to the docstring.

    Parameters
    ----------
    reason : str, optional
        User-friendly reason for deprecating.
    version : str
        Version in which initially marked as deprecated.
    removed_in : str, optional
        Version in which the class or function will be removed.

    Returns
    -------
    decorator

    """
    if version is None:
        raise ValueError('version must be specified.')

    def decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            warning = '%s is deprecated since version %s' % (func.__name__, version)

            if removed_in is not None:
                warning += ' and will be removed in version %s.' % removed_in
            else:
                warning += ' and may be removed in a future version.'

            if reason is not None:
                warning = warning + '  ' + reason

            warnings.warn(warning, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        # Generate Sphinx deprecated directive
        directive = '.. deprecated:: %s' % version

        if reason is not None:
            directive += '\n  %s' % reason

        try:
            functools.update_wrapper(_wrapper, func)

            # Insert directive into original docstring
            _wrapper.__doc__ = _insert_docstring_text(func, directive)
        except AttributeError:
            # __doc__ is not writable in Py27
            pass

        return _wrapper

    return decorator


def experimental(func):
    """Decorate a function or class to designated it as experimental.

    Will raise an `ExperimentalWarning` when used and automatically adds a
    Sphinx '.. warning::' directive to the docstring.

    Parameters
    ----------
    func

    Returns
    -------
    func

    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        warning = (
            '%s is experimental and may be modified or removed without warning.'
            % func.__name__
        )
        warnings.warn(warning, category=ExperimentalWarning, stacklevel=2)
        return func(*args, **kwargs)

    type_ = 'class' if isinstance(func, type) else 'method'
    directive = (
        '.. warning:: This %s is experimental and may be modified or removed without warning.'
        % type_
    )

    try:
        # Insert directive into original docstring
        _wrapper.__doc__ = _insert_docstring_text(func, directive)
    except AttributeError:
        # __doc__ is not writable in Py27
        pass

    return _wrapper


def versionadded(reason=None, version=None):
    """Decorate a function or class to identify the version it was added.

    Automatically adds a Sphinx '.. versionadded::' directive to the docstring.

    Parameters
    ----------
    reason : str, optional
        User-friendly reason for deprecating.
    version : str
        Version in which initially marked as deprecated.

    Returns
    -------
    decorator


    .. versionadded:: 1.5

    """
    if version is None:
        raise ValueError('version must be specified.')

    def decorator(func):
        # Generate Sphinx deprecated directive
        directive = '.. versionadded:: %s' % version

        if reason is not None:
            directive += '\n  %s' % reason

        try:
            # Insert directive into original docstring
            func.__doc__ = _insert_docstring_text(func, directive)
        except AttributeError:
            # __doc__ is not writable in Py27
            pass

        return func

    return decorator


def versionchanged(reason=None, version=None):
    """Decorate a function or class to identify the version it was changed.

    Automatically adds a Sphinx '.. versionchanged::' directive to the docstring.

    Parameters
    ----------
    reason : str, optional
        User-friendly reason for deprecating.
    version : str
        Version in which initially marked as deprecated.

    Returns
    -------
    decorator

    .. versionadded:: 1.5

    """
    if version is None:
        raise ValueError('version must be specified.')

    def decorator(func):
        # Generate Sphinx deprecated directive
        directive = '.. versionchanged:: %s' % version

        if reason is not None:
            directive += '\n  %s' % reason

        try:
            # Insert directive into original docstring
            func.__doc__ = _insert_docstring_text(func, directive)
        except AttributeError:
            # __doc__ is not writable in Py27
            pass

        return func

    return decorator



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


class AuthenticationError(ValueError):
    """A user could not be authenticated."""

    def __init__(self, username=None, msg=None, *args, **kwargs):
        if msg is None:
            if username:
                msg = "Authentication failed for user '%s'." % username
            else:
                msg = "Unable to authenticate the user."

        super(AuthenticationError, self).__init__(msg, *args, **kwargs)


class AuthorizationError(RuntimeError):
    """A user lacks permission to perform an action."""
    pass


class JobTimeoutError(RuntimeError):
    pass



import json

from sasctl.utils.pyml2ds.basic import TreeParser
from .core import EnsembleParser


class XgbTreeParser(TreeParser):
    """Class for parsing xgboost tree."""

    # Generate node dictionary
    def _gen_dict(self):
        pnode = self._node

        self.d[self._node['nodeid']] = self._node
        if 'children' in self._node:
            for child in self._node['children']:
                self._node = child
                self._gen_dict()

        self._node = pnode

    def _not_leaf(self):
        return 'split' in self._node

    def _get_var(self):
        return self._node['split']

    def _go_left(self):
        return self._node['missing'] == self._node['yes']

    def _go_right(self):
        return self._node['missing'] == self._node['no']

    def _left_node(self):
        return self.d[self._node['yes']]

    def _right_node(self):
        return self.d[self._node['no']]

    def _missing_node(self):
        return self.d[self._node['missing']]

    def _split_value(self):
        return self._node['split_condition']

    def _decision_type(self):
        return '<'

    def _leaf_value(self):
        return self._node['leaf']


class XgbParser(EnsembleParser):
    """Class for parsing xgboost model.

    Parameters
    ----------
    booster : xgboost.core.Booster
        Booster of xgboost model.
    objective : {'reg:linear', 'binary:logistic'}
        Xgboost objective function.

    """
    def __init__(self, booster, objective):
        super(XgbParser, self).__init__()

        self._booster = booster
        self._dump = booster.get_dump(dump_format='json')
        self._objective = objective
        self._features = booster.feature_names

        if objective == 'binary:logistic':
            self.out_transform = "1 / (1 + exp(-{0}))"
        elif objective == 'reg:linear':
            pass
        else:
            raise ValueError("Unsupported objective: '%s'.  ""Expected "
                             "'binary:logistic' or 'reg:linear'." % objective)

        self._tree_parser = XgbTreeParser()
    
    def _iter_trees(self):
        for booster_id, tree_json in enumerate(self._dump):
            yield booster_id, json.loads(tree_json)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os

from .folders import Folders
from .service import Service
from sasctl.utils.cli import sasctl_command


class Files(Service):
    """The Files API provides persistence and retrieval of files, such as
    documents, attachments, and reports.

    The file can be associated with the URI of another identifiable object
    (for example, a parentUri). Every file must have an assigned content type
    and name. Files can be retrieved individually by using the file's
    identifier or as a list of files by using a parentUri. Each file has its
    content stream associated with it.  After creation, the metadata that is
    associated with the file or the actual content can be updated. A single
    file can be deleted by using a specific ID. Multiple files can be deleted
    by specifying a parentUri.  A file can be uploaded via raw request or
    multipart form request.
    """

    _SERVICE_ROOT = '/files'

    list_files, get_file, update_file, delete_file = Service._crud_funcs(
        '/files', 'file'
    )

    @sasctl_command('files', 'create')
    def create_file(self, file, folder=None, filename=None, expiration=None):
        """Create a new file on the server by uploading a local file.

        Parameters
        ----------
        file : str or file_like
            Path to the file to upload or a file-like object.
        folder : str or dict, optional
            Name, or, or folder information as returned by :func:`.get_folder`.
        filename : str, optional
            Name to assign to the uploaded file.  Defaults to the filename if `file` is a path, otherwise required.
        expiration : datetime, optional
            A timestamp that indicates when to expire the file.  Defaults to no expiration.

        Returns
        -------
        RestObj
            A dictionary containing the file attributes.

        """
        if isinstance(file, str):
            filename = filename or os.path.splitext(os.path.split(file)[1])[0]

            with open(file, 'rb') as f:
                file = f.read()
        else:
            if filename is None:
                raise ValueError(
                    '`filename` must be specified if `file` is not a path.'
                )

            file = file.read()

        params = {}

        if folder is not None:
            _folder = Folders().get_folder(folder)

            if _folder is None:
                raise ValueError("Folder '%s' could not be found." % folder)

            params['parentFolderUri'] = self.get_link(_folder, 'self')['href']

        if expiration is not None:
            pass
            # TODO: add 'expirationTimeStamp' to params.  Need to determine correct format

        return self.post(
            '/files#multipartUpload', files={filename: file}, params=params
        )

    @sasctl_command('files', 'content')
    def get_file_content(self, file):
        """Download the contents of a file.

        Parameters
        ----------
        file : str or dict, optional
            Name or file information as returned by :func:`get_file`.

        Returns
        -------
        content

        """
        file = self.get_file(file)

        r = self.request_link(file, 'content', format='response')

        content_type = r.headers.get('Content-Type', '')

        if 'text/plain' in content_type:
            return r.text
        if 'application/json' in content_type:
            return r.json()
        if 'application/octet-stream' in content_type:
            return r.content

        return r.text



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Base functionality for all services."""

import logging
import time
import warnings
from urllib.parse import quote

from .. import core
from ..core import HTTPError, PagedItemIterator, sasctl_command
from ..exceptions import JobTimeoutError


class Service(object):  # skipcq PYL-R0205
    """Base class for all services.  Should not be used directly."""

    _SERVICE_ROOT = None

    is_uuid = staticmethod(core.is_uuid)
    get_link = staticmethod(core.get_link)
    request_link = staticmethod(core.request_link)

    log = logging.getLogger(__name__)

    @property
    def _SERVICE_ROOT(self):
        raise NotImplementedError()

    @classmethod
    def is_available(cls):
        """Check if the service is currently available.

        Returns
        -------
        bool

        """
        try:
            response = cls.head('/', format='response')
            return response.status_code == 200
        except HTTPError:
            return False

    @classmethod
    def info(cls):
        """Version and build information for the service.

        Returns
        -------
        RestObj

        """
        return cls.get('/apiMeta')

    @classmethod
    def request(cls, verb, path, session=None, raw=False, format='auto', **kwargs):
        """Send an HTTP request with a session.

        Parameters
        ----------
        verb : str
            A valid HTTP request verb.
        path : str
            Path portion of URL to request.  Assumed to be relative to
            `_SERVICE_ROOT`.
        session : Session, optional
            Defaults to `current_session()`.
        raw : bool
            Deprecated. Whether to return the raw `Response` object.
            Defaults to False.
        format : {'auto', 'response', 'content', 'json', 'text'}
            The format of the return response.  Defaults to `auto`.
            response: the raw `Response` object.
            content: Response.content
            json: Response.json()
            text: Response.text
            auto: `RestObj` constructed from JSON if possible, otherwise same as
                  `text`.
        kwargs : any
            Additional arguments are passed to the session `request` method.

        Returns
        -------

        """
        if path.startswith('/'):
            path = cls._SERVICE_ROOT + path
        else:
            path = cls._SERVICE_ROOT + '/' + path

        return core.request(verb, path, session, format, **kwargs)

    @classmethod
    def get(cls, *args, **kwargs):
        """Send a GET request."""
        try:
            return cls.request('get', *args, **kwargs)
        except HTTPError as e:
            if e.code == 404:
                return None  # Resource not found
            raise e

    @classmethod
    def head(cls, *args, **kwargs):
        """Send a HEAD request."""
        return cls.request('head', *args, **kwargs)

    @classmethod
    def post(cls, *args, **kwargs):
        """Send a POST request."""
        return cls.request('post', *args, **kwargs)

    @classmethod
    def put(cls, *args, **kwargs):
        """Send a PUT request."""
        return cls.request('put', *args, **kwargs)

    @classmethod
    def delete(cls, *args, **kwargs):
        """Send a DELETE request."""
        return cls.request('delete', *args, **kwargs)

    @staticmethod
    def _crud_funcs(
        path, single_term=None, plural_term=None, service_name=None, get_filter=None
    ):
        """Utility method for defining CRUD functions.

        Can be used to define simple functions that perform CRUD operations
        on a REST endpoint.

        Parameters
        ----------
        path : str
            URL path to use for the requests
        single_term : str
            English name of the item being manipulated.
            Defaults to `plural_term`.
        plural_term : str
            English name of the items being manipulated. Defaults to the last
            segment of `path`.
        service_name : str
            Name of the service under which the command will be listed in
            the `sasctl` CLI.  Defaults to `plural_term`.
        get_filter : callable, optional
            A function that accepts an `item` and returns a dictionary
            representing quary parameters to be added to the request for
            filtering.  Defaults to filter=eq(name, item.name).

        Returns
        -------
        functions : tuple
            tuple of `classmethod` instances representating CRUD functions:
            list_items, get_item, update_item, delete_item

        Examples
        --------

        >>> list_spam, get_spam, update_spam, delete_spam = _build_crud_funcs('/spam')

        """
        # Set a default filter
        if get_filter is None:

            def default_filter(item):
                return dict(filter='eq(name, "%s")' % item)

            get_filter = default_filter

        @sasctl_command('list')
        def list_items(cls, filter=None, start=None, limit=None, **kwargs):
            """List all {items} available in the environment.

            Parameters
            ----------
            filter : str, optional
            start : int, optional
                Zero-based index of the first item to return.  Defaults to 0.
            limit : int, optional
                The maximum number of items to return.  Defaults to 20.

            Returns
            -------
            list
                A list of dictionaries containing the {items}.

            Notes
            -----
            See the filtering_ reference for details on the `filter` parameter.

            .. _filtering: https://developer.sas.com/reference/filtering/

            """
            if filter is not None:
                kwargs['filter'] = filter
            if start is not None:
                kwargs['start'] = int(start)
            if limit is not None:
                kwargs['limit'] = int(limit)

            params = '&'.join(
                '%s=%s' % (k, quote(str(v), safe='/(),"')) for k, v in kwargs.items()
            )

            results = cls.get(path, params=params)
            if results is None:
                return []
            if isinstance(results, (list, PagedItemIterator)):
                return results

            return [results]

        @sasctl_command('get')
        def get_item(cls, item, refresh=False):
            """Return a {item} instance.

            Parameters
            ----------
            item : str or dict
                Name, ID, or dictionary representation of the {item}.
            refresh : bool, optional
                Obtain an updated copy of the {item}.

            Returns
            -------
            RestObj or None
                A dictionary containing the {item} attributes or None.

            Notes
            -------
            If `item` is a complete representation of the {item} it will be
            returned unless `refresh` is set.  This prevents unnecessary REST
            calls when data is already available on the client.

            """
            # If the input already appears to be the requested object just
            # return it, unless a refresh of the data was explicitly requested.
            if isinstance(item, dict) and all(k in item for k in ('id', 'name')):
                if refresh:
                    item = item['id']
                else:
                    return item

            if cls.is_uuid(item):
                return cls.get(path + '/{id}'.format(id=item))
            results = list_items(cls, **get_filter(item))

            match = None
            for result in results:
                if result['name'] == str(item):
                    # The first result that matches on name should be stored.
                    # Will be returned after determining that there aren't additional
                    # matches
                    if match is None:
                        # Make a request for the specific object so that ETag
                        # is included, allowing updates.
                        if cls.get_link(result, 'self'):
                            match = cls.request_link(result, 'self')
                        else:
                            id_ = result.get('id', result['name'])
                            match = cls.get(path + '/{id}'.format(id=id_))
                    # We already found a match so this is a duplicate.  Warn the user so they know
                    # the item returned may not be the one they were expecting.
                    else:
                        warnings.warn(
                            "Multiple items found with name '%s'.  Only the first result is returned."
                            % item
                        )
                        break

            return match

        @sasctl_command('update')
        def update_item(cls, item):
            """Update a {item} instance.

            Parameters
            ----------
            item : dict

            Returns
            -------
            None

            """
            headers = getattr(item, '_headers', None)
            if headers is None or headers.get('etag') is None:
                raise ValueError('Could not find ETag for update of %s.' % item)

            id_ = getattr(item, 'id', None)
            if id_ is None:
                raise ValueError(
                    'Could not find property `id` for update of %s.' % item
                )

            headers = {
                'If-Match': item._headers.get('etag'),
                'Content-Type': item._headers.get('content-type'),
            }

            return cls.put(path + '/%s' % id_, json=item, headers=headers)

        @sasctl_command('delete')
        def delete_item(cls, item):
            """Delete a {item} instance.

            Parameters
            ----------
            item

            Returns
            -------
            None

            """
            item_name = str(item)

            # Try to find the item if the id can't be found
            if not (isinstance(item, dict) and 'id' in item):
                item = get_item(cls, item)
                if item is None:
                    cls.log.info("Object '%s' not found.  Skipping delete.", item_name)
                    return

            if isinstance(item, dict) and 'id' in item:
                item = item['id']

            if cls.is_uuid(item):
                response = cls.delete(path + '/{id}'.format(id=item))
                # Response generally seems to be an empty string.  If so, just return None
                # BUT, if the service provides an actual response, return it.
                if response:
                    return response
                return
            raise ValueError("Unrecognized id '%s'" % item)

        # Pull object name from path if unspecified (many paths end in
        # nouns like /folders or /repositories).
        plural_term = plural_term or str(path).split('/')[-1]
        single_term = single_term or plural_term
        service_name = service_name or plural_term
        service_name = service_name.replace(' ', '_')

        for func in [list_items, get_item, update_item, delete_item]:
            func.__doc__ = func.__doc__.format(item=single_term, items=plural_term)
            func._cli_service = service_name

            prefix = func.__name__.split('_')[0] + '_'
            suffix = plural_term if prefix == 'list_' else single_term
            func.__name__ = prefix + suffix

        return [
            classmethod(f) for f in (list_items, get_item, update_item, delete_item)
        ]

    # Compatibility with Python 2.7 requires *args to be after key-words
    # arguments.
    # skipcq: PYL-W1113
    def _get_rel(self, item, rel, func=None, filter=None, *args):
        """Get `item` and request a link.

        Parameters
        ----------
        item : str or dict
        rel : str
        func : function, optional
            Callable that takes (item, *args) and returns a RestObj of `item`
        filter : str, optional

        args : any
            Passed to `func`

        Returns
        -------
        list

        """
        if func is not None:
            obj = func(item, *args)

        if obj is None:
            return

        params = 'filter={}'.format(filter) if filter is not None else {}

        resources = self.request_link(obj, rel, params=params)

        if isinstance(resources, (list, PagedItemIterator)):
            return resources

        return [resources]

    @classmethod
    def _monitor_job(cls, job, max_retries=60):
        """Continually poll a job until it reaches the desired status.

        Parameters
        ----------
        job : dict
            Dictionary representation of a currently execution job
        max_retries : int
            Maximum number of ties to refresh `job` before failing.

        Returns
        -------
        job

        Raises
        ------
        TimeoutError
            `max_retries` reached with a successful status check

        """

        def completed(job):
            return job['state'].lower() in ('completed', 'failed')

        retries = 0

        if cls.get_link(job, 'self') is None:
            raise ValueError("Link 'self' not found on %s" % job)

        # TODO: Log
        while not completed(job) and retries < max_retries:
            time.sleep(0.5)
            retries += 1
            job = cls.request_link(job, 'self')

        if completed(job):
            return job
        raise JobTimeoutError('Timeout while waiting on job %s' % job)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

__version__ = '1.6.1'
__author__ = 'SAS'
__credits__ = ['Yi Jian Ching', 'Lucas De Paula', 'James Kochuba', 'Peter Tobac',
               'Chris Toth', 'Jon Walker', 'Scott Lindauer']
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright © 2019, SAS Institute Inc., ' \
                'Cary, NC, USA.  All Rights Reserved.'

import logging


from .core import current_session, delete, get, get_link, platform_version, post, put, request_link
from .core import RestObj, Session, HTTPError
from .tasks import register_model, publish_model, update_model_performance


# Prevent package from emitting log records unless consuming
# application configures logging.
logging.getLogger(__name__).addHandler(logging.NullHandler())



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .astore import create_package, create_package_from_astore





import abc
import unicodedata


class TreeParser(metaclass=abc.ABCMeta):
    """Abstract class for parsing decision tree.

    Attributes
    ----------
    d : dict
        Dictionary for storing node hierarchy. Not used in all models.
    out_transform : string
        Output transformation for generated value. For example, if logreg is
        used: 1 / (1 + exp(-{0})), where {0} stands for resulting gbvalue.
    out_var_name : string
        Name used for output variable.

    """

    def __init__(self):
        self.d = {}
        self._indent = 4
        self._depth = -1
        self._tree_id = None
        self._node = None
        self._root = None

    def init(self, root, tree_id=0):
        """Custom init method. Need to be called before using TreeParser.

        Attributes
        ----------
        root : node
            Tree root node.
        tree_id : int
            Id of current tree.

        """
        self._root = root
        self._node = root
        self._tree_id = tree_id
        self._depth = -1
        self._indent = 4

    def _gen_dict(self):
        self.d = {}

    def _get_indent(self):
        return " " * self._indent * self._depth

    @abc.abstractmethod
    def _not_leaf(self):
        pass

    @abc.abstractmethod
    def _get_var(self):
        pass

    @abc.abstractmethod
    def _go_left(self):
        pass

    @abc.abstractmethod
    def _go_right(self):
        pass

    @abc.abstractmethod
    def _left_node(self):
        pass

    @abc.abstractmethod
    def _right_node(self):
        pass

    @abc.abstractmethod
    def _missing_node(self):
        pass

    @abc.abstractmethod
    def _split_value(self):
        pass

    @abc.abstractmethod
    def _decision_type(self):
        pass

    @abc.abstractmethod
    def _leaf_value(self):
        pass

    @staticmethod
    def _remove_diacritic(input):
        output = unicodedata.normalize('NFKD', input).encode('ASCII', 'ignore').decode()

        return output

    def parse_node(self, f, node=None):
        """Recursively parse tree node and write generated SAS code to file.

        Attributes
        ----------
        f : file object
            Open file for writing output SAS code.
        node: node
            Tree node to process.

        """

        pnode = self._node
        if node is not None:
            self._node = node
        else:
            self._node = self._root

        if self._node == self._root:
            self.d = {}
            self._gen_dict()

        self._depth += 1

        if self._not_leaf():
            var = self._get_var()[:32]
            var = self._remove_diacritic(var)

            split_value = self._split_value()

            cond = ""
            if self._go_left():
                cond = "missing({}) or ".format(var)
            elif self._go_right():
                cond = "not missing({}) and ".format(var)
            else:
                f.write(self._get_indent() + "if (missing(%s)) then do;\n" % var)
                self.parse_node(f, node=self._missing_node())
                f.write(self._get_indent() + "end;\n")

            f.write(
                self._get_indent()
                + "if ({}{} {} {}) then do;\n".format(
                    cond, var, self._decision_type(), split_value
                )
            )
            self.parse_node(f, node=self._left_node())
            f.write(self._get_indent() + "end;\n")
            f.write(self._get_indent() + "else do;\n")
            self.parse_node(f, node=self._right_node())
            f.write(self._get_indent() + "end;\n")
        else:
            leaf_value = self._leaf_value()

            f.write(
                self._get_indent() + "treeValue%s = %s;\n" % (self._tree_id, leaf_value)
            )

        self._node = pnode
        self._depth -= 1



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .service import Service


class Reports(Service):
    """Creates, reads, updates, and deletes reports, report states, and content.

    See Also
    --------
    `REST Documentation <https://developer.sas.com/apis/rest/Visualization/#
    reports>`_

    """

    _SERVICE_ROOT = 'reports'

    list_reports, get_report, _, _ = Service._crud_funcs('/reports', 'report')

    @classmethod
    def get_visual_elements(cls, report):
        """Get the visual components of a report.

        Returned components may be visualized by rendering with the
        `report_images.get_images` method.

        Parameters
        ----------
        report : str or dict
            The name or id of the report, or a dictionary representation of the
            report.

        Returns
        -------
        List[RestObj]
            List of metadata about each element.

        """
        report = cls.get_report(report)
        elements = cls.request_link(report, 'contentVisualElements')

        # Despite being "visual" not all elements can be rendered by
        # report_images service.  Only return renderable elements.
        return [e for e in elements if e.type not in ('Table')]






from sasctl.utils.pyml2ds.basic import TreeParser
from .core import EnsembleParser


class PmmlTreeParser(TreeParser):
    """Class for parsing pmml gradient boosting tree."""

    def _not_leaf(self):
        return self._node.get('defaultChild')

    def _get_var(self):
        return self._node.find('Node').find('SimplePredicate').get('field')

    def _go_left(self):
        return self._node.find('Node').get('id') == self._node.get(
            'defaultChild')

    def _go_right(self):
        return not self._node.find('Node').get('id') == self._node.get(
            'defaultChild')

    def _left_node(self):
        return self._node.findall('Node')[0]

    def _right_node(self):
        return self._node.findall('Node')[1]

    def _missing_node(self):
        return None

    def _split_value(self):
        return self._node.find('Node').find('SimplePredicate').get('value')

    def _decision_type(self):
        ops = {
            'lessThan': '<',
            'lessOrEqual': '<=',
            'greaterThan': '>',
            'greaterOrEqual': '>='
        }

        return ops[self._node.find('Node').find('SimplePredicate').get('operator')]

    def _leaf_value(self):
        return self._node.get('score')


class PmmlParser(EnsembleParser):
    """Class for parsing pmml gradient boosting models.

    Parameters
    ----------
    tree_root : etree.Element
        Root node of pmml gradient boosting forest.

    """
    def __init__(self, tree_root):
        super(PmmlParser, self).__init__()

        self._tree_root = tree_root
        for elem in tree_root.getiterator():
            if hasattr(elem.tag, 'find'):
                i = elem.tag.find('}')
                if i >= 0:
                    elem.tag = elem.tag[i + 1:]

        self._forest = tree_root.find('MiningModel/Segmentation')[0].find('MiningModel')

        rescaleConstant = self._forest.find('Targets/Target').get('rescaleConstant')
        self.out_transform = "1 / (1 + exp(-{}))".format("({0} + " + "{})".format(rescaleConstant))

        self._tree_parser = PmmlTreeParser()

    def _iter_trees(self):
        for booster_id, tree_elem in enumerate(self._forest.find('Segmentation')):
            yield booster_id, tree_elem.find('TreeModel/Node')



from sasctl.utils.pyml2ds.basic import TreeParser
from .core import EnsembleParser


class LightgbmTreeParser(TreeParser):
    """Class for parsing lightgbm tree.
    """

    def _not_leaf(self):
        return 'split_feature' in self._node

    def _get_var(self):
        return self._features[self._node['split_feature']]

    def _go_left(self):
        return self._node['default_left']

    def _go_right(self):
        return not self._node['default_left']

    def _left_node(self):
        return self._node['left_child']

    def _right_node(self):
        return self._node['right_child']

    def _missing_node(self):
        return None

    def _split_value(self):
        return self._node['threshold']

    def _decision_type(self):
        return self._node['decision_type']

    def _leaf_value(self):
        return self._node['leaf_value']


class LightgbmParser(EnsembleParser):
    """Class for parsing lightgbm model.

    Parameters
    ----------
    booster : lightgbm.basic.Booster
        Booster of lightgbm model.

    """
    def __init__(self, booster):
        super(LightgbmParser, self).__init__()

        self._booster = booster
        self._dump = booster.dump_model()

        objective = self._dump.get('objective')

        if objective != 'binary sigmoid:1':
            raise ValueError("Only binary sigmoid objective function is "
                             "currently supported. Received '%s'."
                             % objective)

        self._features = self._dump['feature_names']
        self.out_transform = "1 / (1 + exp(-{0}))"

        self._tree_parser = LightgbmTreeParser()
        self._tree_parser._features = self._features

    def _iter_trees(self):
        for tree in self._dump['tree_info']:
            yield tree['tree_index'], tree['tree_structure']



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os
import re

from setuptools import setup, find_packages


def read_dunder(name):
    with open(os.path.join('src', 'sasctl', '__init__.py')) as f:
        for line in f.readlines():
            match = re.search(r'(?<=^__{}__ = [\'"]).*\b'.format(name), line)
            if match:
                return match.group(0)
        raise RuntimeError('Unable to find __%s__ in __init__.py' % name)


def get_file(filename):
    with open(filename, 'r') as f:
        return f.read()


setup(
    name='sasctl',
    description='SAS Viya Python Client',
    long_description=get_file('README.md'),
    long_description_content_type='text/markdown',
    version=read_dunder('version'),
    author=read_dunder('author'),
    license='Apache v2.0',
    url='https://github.com/sassoftware/python-sasctl/',
    project_urls={
        'Bug Tracker': 'https://github.com/sassoftware/python-sasctl/issues',
        'Documentation': 'https://sassoftware.github.io/python-sasctl/',
        'Source Code': 'https://github.com/sassoftware/python-sasctl'
    },
    include_package_data=True,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires='>=3.5',
    install_requires=[
        'pandas',
        'scikit-learn',
        'requests',
        'pyyaml',
        'packaging'
    ],
    extras_require={
        'swat': ['swat'],
        'kerberos': ['kerberos ; platform_system != "Windows"',
                     'winkerberos ; platform_system == "Windows"'],
        'all': ['swat',
                'kerberos ; platform_system != "Windows"',
                'winkerberos ; platform_system == "Windows"'],
    },
    entry_points={'console_scripts': ['sasctl = sasctl.utils.cli:main']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Operating System :: OS Independent'
    ]
)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import random
import string

from .decorators import versionadded


def installed_packages():
    """List Python packages installed in the current environment.

    Returns
    -------

    Notes
    -----
    Uses pip freeze functionality so pip module must be present.

    """
    try:
        from pip._internal.operations import freeze
    except ImportError:
        try:
            from pip.operations import freeze
        except ImportError:
            freeze = None

    if freeze is not None:
        return list(freeze.freeze())


@versionadded(version='1.5.1')
def random_string(length):
    """Generates a random alpha-numeric string of a given length.

    Parameters
    ----------
    length : int
        The length of the generate string.

    Returns
    -------
    str

    """

    # random.choices() wasn't added until Python 3.6, so repeatedly call .choice() instead
    chars = string.ascii_letters + string.digits
    return ''.join(random.choice(chars) for _ in range(length))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from ..core import PagedItemIterator
from .service import Service


class DataSources(Service):
    """Enables retrieval of data source metadata.

    The Data Sources API works in concert with the Data Tables and Row Sets
    APIs to navigate, reference, and retrieve data in the SAS Viya ecosystem.
    The Data Sources API enables retrieval of metadata for data sources and
    linking to their respective tables.

    """

    _SERVICE_ROOT = '/dataSources'

    list_providers, _, _, _ = Service._crud_funcs('/providers', 'provider')

    def get_provider(self, provider, refresh=False):
        """Returns a provider instance.

        Parameters
        ----------
        provider : str or dict
            Name, ID, or dictionary representation of the provider.
        refresh : bool, optional
            Obtain an updated copy of the {item}.

        Returns
        -------
        RestObj or None
            A dictionary containing the provider attributes or None.

        Notes
        -------
        If `provider` is a complete representation of the provider it will be
        returned unless `refresh` is set.  This prevents unnecessary REST calls
        when data is already available on the client.

        """
        if isinstance(provider, dict) and 'id' in provider:
            if refresh:
                provider = provider['id']
            else:
                return provider

        return self.get('/providers/{id}'.format(id=provider))

    def list_sources(self, provider):
        """List all data sources available for a provider.

        Parameters
        ----------
        provider : str or dict
            Name, ID, or dictionary representation of the provider.

        Returns
        -------
        list
            A collection of :class:`.RestObj` instances.

        """

        provider = self.get_provider(provider)

        sources = self.request_link(provider, 'dataSources')
        if isinstance(sources, (list, PagedItemIterator)):
            return sources
        return [sources]

    def get_source(self, provider, source):
        """Returns a data source belonging to a given provider.

        Parameters
        ----------
        provider : str or dict
            Name, ID, or dictionary representation of the provider.
        source : str
            Name or id of the data source

        Returns
        -------
        RestObj or None
            A dictionary containing the data source attributes or None.

        """
        if isinstance(source, dict) and 'providerId' in source:
            return source

        sources = self.list_sources(provider)

        for s in sources:
            if source in (s.name, s.id):
                return s

    def list_caslibs(self, source='cas-shared-default', filter=None):
        """Get all caslibs registered with the given CAS server.

        Parameters
        ----------
        source : str, optional
            Name of the CAS server.  Defaults to `cas-shared-default`.
        filter : str, optional

        Returns
        -------
        list
            A collection of :class:`.RestObj` instances.

        Notes
        -----
        See the filtering_ reference for details on the `filter` parameter.

        .. _filtering: https://developer.sas.com/reference/filtering/

        """
        source = self.get_source('cas', source)

        params = 'filter={}'.format(filter) if filter is not None else {}
        result = self.request_link(source, 'children', params=params)

        return result if isinstance(result, (list, PagedItemIterator)) else [result]

    def get_caslib(self, name, source=None):
        """Get a caslib by name.

        Parameters
        ----------
        name : str
            Name of the caslib
        source : str, optional
            Name of the CAS server.  Defaults to `cas-shared-default`.

        Returns
        -------
        RestObj

        """
        source = source or 'cas-shared-default'
        caslibs = self.list_caslibs(source, filter='eq(name, "%s")' % name)

        if caslibs:
            return caslibs[0]

    def list_tables(self, caslib, filter=None):
        """List tables available in a caslib.

        Parameters
        ----------
        caslib : str or dict
            Name, ID, or dictionary representation of the caslib.
        filter : str, optional

        Returns
        -------
        list
            A collection of :class:`.RestObj` instances.

        Notes
        -----
        See the filtering_ reference for details on the `filter` parameter.

        .. _filtering: https://developer.sas.com/reference/filtering/

        """
        if not self.get_link(caslib, 'tables'):
            caslib = self.get_caslib(caslib)

        params = 'filter={}'.format(filter) if filter is not None else {}
        result = self.request_link(caslib, 'tables', params=params)

        return result if isinstance(result, (list, PagedItemIterator)) else [result]

    def get_table(self, name, caslib):
        """Get metadata for a CAS table.

        Parameters
        ----------
        name : str
            Name of the table
        caslib : str or dict
            Name, ID, or dictionary representation of the caslib.

        Returns
        -------
        RestObj

        """
        tables = self.list_tables(caslib, filter='eq(name, "%s")' % name)

        if tables:
            return tables[0]



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright © 2019, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .service import Service
from ..core import uri_as_str


class TextCategorization(Service):
    """Categorizes natural language text documents according to a prebuilt or
    user-defined model.
    """

    _SERVICE_ROOT = '/categorization'

    def categorize(
        self,
        documents,
        model,
        caslib=None,
        id_column=None,
        text_column=None,
        description=None,
        output_postfix=None,
    ):
        """

        Parameters
        ----------
        documents : str or dict or list_like:
            Documents to parse.  May be either the URI to a CAS table where the
            documents are currently stored, or an iterable of strings
            containing the documents' text.
        model : str or dict
            URI of a CAS table that contains one or more category model
            binaries.
        caslib : str or dict, optional
            URI of a caslib in which the documents will be stored.  Required if
            `documents` is a list of strings.
        id_column : str, optional
            The column in `documents` that contains a unique id for each
            document.  Required if `documents` is a CAS table URI.
        text_column : str, optional
            The column in `documents` that contains the document text to
            categorize.
            Required if `documents` is a CAS table URI.
        description : str, optional
            Description to add to the text categorization job.
        output_postfix : str, optional
            Text to be added to the end of all output table names.

        Returns
        -------
        RestObj
            The submitted job

        See Also
        --------
        :meth:`cas_management.get_caslib <.CASManagement.get_caslib>`
        :meth:`cas_management.get_table <.CASManagement.get_table>`

        """
        if documents is None:
            raise TypeError('`documents` cannot be None.')

        url = '/jobs'

        if isinstance(documents, (dict, str)):
            # Input is caslib
            data = {
                'inputUri': uri_as_str(documents),
                'documentIdVariable': id_column,
                'textVariable': text_column,
                'version': 1,
            }
            headers = {
                'Content-Type': 'application/vnd.sas.text.categorization.job.request+json',
                'Accept': 'application/vnd.sas.text.categorization.job+json',
            }
        else:
            # Input is inline documents
            data = {
                'caslibUri': uri_as_str(caslib),
                'documents': documents,
                'version': 1,
            }
            url += '#data'
            headers = {
                'Content-Type': 'application/vnd.sas.text.categorization.job.request.documents+json',
                'Accept': 'application/vnd.sas.text.categorization.job+json',
            }

        data.update(
            {
                'description': description,
                'modelUri': uri_as_str(model),
                'outputTableNamePostfix': output_postfix,
            }
        )

        # Optional fields are not ignored if None. Explicitly remove before sending
        for k in list(data.keys()):
            if data[k] is None:
                del data[k]

        return self.post(url, json=data, headers=headers)



# Copyright (c) 2020, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


# %%
from pathlib import Path

import pickle
import gzip
import codecs

# %%
class PickleModel():
    
    def pickleTrainedModel(self, trainedModel, modelPrefix, pPath=Path.cwd(), isH2OModel=False, isBinaryModel=False, isBinaryString=False):
        '''
        Write trained model to a binary pickle file, H2O MOJO file, or a binary string object. 
        
        The following files are generated by this function:
		* '*.pickle'
			Binary pickle file containing a trained model.
        * '*.mojo'
            Archived H2O.ai MOJO file containing a trained model.
        
        Parameters
        ---------------
        trainedModel : model or string or Path
            For non-H2O models, this argument contains the model variable. Otherwise,
            this should be the file path of the MOJO file.
        modelPrefix : string
            Variable name for the model to be displayed in SAS Open Model Manager 
            (i.e. hmeqClassTree + [Score.py || .pickle]).
        pPath : string, optional
            File location for the output pickle file. Default is the current
            working directory.
        isH2OModel : boolean, optional
            Sets whether the model file is an H2O.ai MOJO file. If set as True, 
            the MOJO file will be gzipped before uploading to SAS Model Manager.
            The default value is False.
        isBinaryModel : boolean, optional
            Sets whether the H2O model provided is a binary model or a MOJO model. By default False.
        isBinaryString : boolean, optional
            Sets whether the model is to be set as a binary string instead of a pickle file. By default False.
            	
        Returns
        -------
        binaryString : binary string
            When the isBinaryString flag is set to True, return a binary string representation of the model instead
            of a pickle or MOJO file.

        '''
        
        if isBinaryString:
            binaryString = codecs.encode(pickle.dumps(trainedModel), 'base64').decode()
            return binaryString
        else:
            # For non-H2O models, pickle the model object
            if not isH2OModel:
                with open(Path(pPath) / (modelPrefix + '.pickle'), 'wb') as pFile:
                    pickle.dump(trainedModel, pFile)
                print('Model {} was successfully pickled and saved to {}.'.format(modelPrefix, Path(pPath) / (modelPrefix + '.pickle')))
            # For H2O models that are binary files, rename the binary file as a pickle file
            elif isBinaryModel:
                binaryFile = Path(pPath) / modelPrefix
                binaryFile.rename(binaryFile.with_suffix('.pickle'))
            # For H2O models in the MOJO format, gzip the model file and rename it with a .MOJO extension
            else:
                with open(Path(trainedModel), 'rb') as fileIn, gzip.open(Path(pPath) / (modelPrefix + '.mojo'), 'wb') as fileOut:
                    fileOut.writelines(fileIn)
                print('MOJO model {} was successfully gzipped and saved to {}.'.format(modelPrefix, Path(pPath) / (modelPrefix + '.mojo')))


import abc


class EnsembleParser(metaclass=abc.ABCMeta):
    """Abstract class for parsing decision tree ensembles.

    Attributes
    ----------
    out_transform : string
        Output transformation for generated value. For example, if logreg is
        used: 1 / (1 + exp(-{0})), where {0} stands for resulting gbvalue.
    out_var_name : string
        Name used for output variable.

    """

    def __init__(self, out_transform="{0}", out_var_name="P_TARGET"):
        self.out_transform = out_transform
        self.out_var_name = out_var_name

    @abc.abstractmethod
    def _iter_trees(self):
        pass

    @classmethod
    def _aggregate(cls, booster_count):
        return "treeValue = sum({});\n".format(
            ', '.join("treeValue%d" % i for i in range(booster_count))
        )

    def translate(self, file):
        """Translate a gradient boosting model and write SAS scoring code to
        a file.

        Attributes
        ----------
        file : file object
            Open file for writing output SAS code.

        """
        for booster_id, tree in self._iter_trees():
            file.write("/* Parsing tree {}*/\n".format(booster_id))

            self._tree_parser.init(tree, booster_id)
            self._tree_parser.parse_node(file)

            file.write("\n")

        file.write("/* Getting target probability */\n")
        file.write(self._aggregate(booster_id + 1))
        file.write(
            "{} = {};\n".format(
                self.out_var_name, self.out_transform.format("treeValue")
            )
        )
